import shutil
import sys
import tomllib

from wox.constants import DOT_WOX, ExitCode
from wox.main import main
from wox.utils import auto_clean_logs
from wox.wox_exceptions import (
    ConfigFileNotFoundError,
    MissingWoxKeywordError,
    MissingWoxSectionError,
    NoTaskError,
    TaskNotFoundError,
    WoxException
)
from wox.wox_steward import WoxSteward


def cli_controller(cli_options):
    """
        Processes cli options to make them compatible with the execution
        of wox runner, then starts the wox session.

        Parameters
        ----------
        cli_options : dict
            Contains options provided by the user in command-line.

        Returns
        -------
        exit_code : int

        Raises
        ------
        Exception
            Any exception not handled by wox.
    """

    wox_steward = WoxSteward()
    wox_steward.dot_wox_setup()

    if cli_options.get('purge'):
        wox_steward.purge_wox()
        print('.wox directory successfully purged.')
        return ExitCode.SUCCESS.value

    try:
        return main(cli_options, wox_steward.wox_data)
    except (
        ConfigFileNotFoundError,
        MissingWoxSectionError,
        MissingWoxKeywordError,
        NoTaskError,
        TaskNotFoundError,
        tomllib.TOMLDecodeError
    ) as e:
        if isinstance(e, WoxException):
            print(e)
        else:
            print(f'{e.__class__.__name__}: {e}')

        if 'pytest' not in sys.modules:
            auto_clean_logs()

        # handling case when wox is executed naively for the first time
        if (
            isinstance(e, ConfigFileNotFoundError) or \
            isinstance(e, MissingWoxSectionError)
        ):
            logs_dir = DOT_WOX / '.logs'
            if not any(path.is_file() for path in logs_dir.rglob('*')):
                shutil.rmtree(DOT_WOX)

        return ExitCode.CONFIG_ERROR.value
    except KeyboardInterrupt:
        print('Wox session aborted by user.')
        return ExitCode.INTERRUPTED.value
    except Exception:
        if 'pytest' not in sys.modules:
            auto_clean_logs()
        raise



# coverage run -m pytest -rA tests/test_subprocessor.py --no-header --html=./reports/pytest/report.html --self-contained-html && coverage html #noqa E501

import pytest

import wox.python_discoverer as python_discoverer
import wox.venv_management as venv_manager

from wox.subprocessor import sanitize_command
from wox.wox_exceptions import CommandNotFoundError, PythonInterpreterCallError


# python -B -m pytest -rA -k "test_sanitize_command"
def test_sanitize_command(monkeypatch, tmp_path):
    """
        The python interpreter must be added to PATH environment variable, ensuring
        the system will use the right python interpreter.


        If the command is 'python', the python interpreter path provided by the virtual
        environment must be casted into string for compatibility with shutil.which().
    """

    env = 'py311'

    python_interpreter = python_discoverer.python_discovery(env)

    env_path = tmp_path / env

    venv_manager.create_venv(env_path, python_interpreter)

    venv_python_interpreter = python_discoverer.locate_venv_python_exe(env_path)

    assert venv_python_interpreter.path.exists()

    def patched_which(cmd = None, path = None):
        patched_which.cmd = cmd
        patched_which.path = path
        return venv_python_interpreter

    monkeypatch.setattr('shutil.which', patched_which)

    sanitize_command('python -V', venv_python_interpreter.path)

    assert isinstance(patched_which.cmd, str)
    assert str(venv_python_interpreter.path.parent) in patched_which.path

# python -B -m pytest -rA -k "test_python_interpreter_call_error"
def test_python_interpreter_call_error():
    """
        PythonInterpreterCallError must raise if the command
        would run the python REPL.
    """

    with pytest.raises(PythonInterpreterCallError):
        sanitize_command('python')

# python -B -m pytest -rA -k "test_command_not_found_error"
def test_command_not_found_error():
    """
        CommandNotFoundError must raise if the command doesn't exist.
    """

    with pytest.raises(CommandNotFoundError):
        sanitize_command('cowsa')



# coverage run -m pytest -rA tests/test_python_discoverer.py --no-header --html=./reports/pytest/report.html --self-contained-html && coverage html #noqa E501

import pytest

import wox.python_discoverer as python_discoverer


class TestGetInterpreterVersion:
    """
        get_interpreter_version() must return a version number
        in the major.minor format, extracted from virtual environment
        names following the 'py<majorminor>' convention (e.g. py311).
    """

    test_get_interpreter_version_cases = [
        {
            'input': 'py30',
            'expected': '3.0'
        },
        {
            'input': 'py31',
            'expected': '3.1'
        },
        {
            'input': 'py310',
            'expected': '3.10'
        },
        {
            'input': 'py311',
            'expected': '3.11'
        },
        {
            'input': 'py3100',
            'expected': '3.100'
        },
        {
            'input': 'py31000',
            'expected': '3.1000'
        },
    ]

    # python -B -m pytest -rA -k "test_get_interpreter_version"
    @pytest.mark.parametrize(
        'test_case_',
        test_get_interpreter_version_cases,
        ids = [case['input'] for case in test_get_interpreter_version_cases]
    )
    def test_get_interpreter_version(self, test_case_):

        env = test_case_['input']

        version = python_discoverer.get_interpreter_version(env)

        assert version == test_case_['expected']

@pytest.mark.parametrize(
    'env',
    [
        'py311',
        '.venv'
    ]
)
# python -B -m pytest -rA -k "test_python_discovery"
def test_python_discovery(env):
    """
        Wox must discover the required Python version if
        a virtual environment name follows the 'py<majorminor>' convention.

        Otherwise, Wox must discover the Python interpreter available
        in the current environment or provided by the system.
    """

    python_interpreter = python_discoverer.python_discovery(env)

    assert python_interpreter.path.exists()

class TestGetPythonVersionFromPyvenvCfg:
    """
        get_python_version_from_pyvenv_cfg() must return a Python version
        using the major.minor.fix scheme.
    """

    test_get_python_version_from_pyvenv_cfg_cases = [
        {
            'input': '3.12.1-rc1',
            'expected': '3.12.1'
        },
        {
            'input': '3.11.7.final.0',
            'expected': '3.11.7'
        },
        {
            'input': '3.12.11a7',
            'expected': '3.12.11'
        },
        {
            'input': '3.100.1000-beta-3.14.0-final',
            'expected': '3.100.1000'
        }
    ]

    # python -B -m pytest -rA -k "test_get_python_version_from_pyvenv_cfg"
    @pytest.mark.parametrize(
        'test_case_',
        test_get_python_version_from_pyvenv_cfg_cases
    )
    def test_get_python_version_from_pyvenv_cfg(self, tmp_path, test_case_):

        pyvenv_cfg = tmp_path / 'pyvenv.cfg'

        pyvenv_cfg.write_text(f"version = {test_case_['input']}")

        version = python_discoverer.get_python_version_from_pyvenv_cfg(tmp_path)

        assert version == test_case_['expected']



from wox.logger.logger_setup import wox_logger
from wox.utils import sort_set
from wox.wox_exceptions import (
    MissingWoxKeywordError,
    NoTaskError,
    TaskNotFoundError
)


def configuration_creation(env, settings, envlist):
    """
        Creates the configuration to execute task.

        Parameters
        ----------
        env : str
            The name of the virtual environment.
        settings : dict
            Settings exctracted from the configuration .toml file.
        envlist : list
            List of virtual environments used during the wox session.

        Returns
        -------
        dict

        Notes
        -----
        When an env has specific settings in its subsection, they are
        merged with the settings from the main section to create a
        complete configuration.
        This is done with an union between the two dict. When two keys
        are identical, the key from the first dict is kept.
        The union operator was introduced in Python 3.9.
    """

    wox_logger.debug('Creation of configuration for the current task.')

    base_config = {
        key: value
        for key, value in settings.items()
        if key not in envlist and not isinstance(value, dict)
    }

    env_config = settings.get(env, {})

    config = base_config | env_config

    wox_logger.debug(f'Configuration: {config}')

    return config

def session_setup(cli_options, wox_data):
    """
        Resolves session configuration and selected tasks based on command-line options
        and the content of the sections of the .toml file.

        Notes
        -----
        Command-line options take priority over configuration values.

        If both ``--task`` and ``--exclude-task`` are present among CLI options,
        ``--task`` has priority over ``--exclude-task``, without denying it:
        a task appearing in both ``--task`` and ``--exclude-task`` is included.
        All other tasks listed in ``--exclude-task`` are excluded.

        Parameters
        ----------
        cli_options : dict
            Options provided on the command line.
        wox_data : dict
            Information provided by the .toml file.

        Returns
        -------
        wox_configuration : dict
            The configuration extracted from the ``wox_configuration`` section
            of the .toml file.
        wox_data : dict
            Dictionnary with the tasks declared in the .toml file,  filtered
            according to wox configuration and command-line options.

        Raises
        ------
        MissingWoxKeywordError
            Required wox keyword missing in .toml file.
        NoTaskError
            No tasks left with use of --task and --exclude-task.
        TaskNotFoundError
            The task is not present in the .toml file.
    """

    try:
        wox_configuration = wox_data.pop('wox_configuration')
    except KeyError as e:
        raise MissingWoxKeywordError(e)

    all_tasks = set(wox_data.keys())
    if not all_tasks:
        raise NoTaskError
    cli_task = set(cli_options.get('task') or [])
    cli_exclude = set(cli_options.get('exclude_task') or [])
    default = set(wox_configuration.get('wox_session_default') or [])

    if cli_task:
        missing_tasks = cli_task - all_tasks
        sorted_missing_tasks = sort_set(missing_tasks, cli_options['task'])
        if missing_tasks:
            raise TaskNotFoundError(
                ', '.join([f"'{task}'" for task in sorted_missing_tasks])
            )
        session_tasks = cli_task
    elif default:
        session_tasks = default
    else:
        session_tasks = all_tasks

    if cli_exclude:
        missing_tasks = cli_exclude - all_tasks
        sorted_missing_tasks = sort_set(missing_tasks, cli_options['exclude_task'])
        if missing_tasks:
            raise TaskNotFoundError(
                ', '.join([f"'{task}'" for task in sorted_missing_tasks])
            )
        session_tasks.difference_update(cli_exclude)
        if not session_tasks:
            raise NoTaskError

    sorted_tasks = sort_set(session_tasks, wox_data.keys())

    wox_data = {task: wox_data[task] for task in sorted_tasks}

    return wox_configuration, wox_data



"""
    This module contains (trouver mieux, comme rassembler) all errors that may occur during
    the naive discovery of Python happening in subprocessor call from python_discoverer.py.
"""

from wox.utils import get_python_version_from_command
from wox.wox_exceptions import PythonBuildNotFoundError, PythonInterpreterNotFoundError


def detect_python_discovery_errors(command):
    """
        Detects errors upon naive python discovery.

        Parameters
        ----------
        command: str
            Command executed by subprocess.

        Raises
        ------
        PythonInterpreterNotFoundError
            The Python interpreter for the specified version cannot be found.
        PythonBuildNotFoundError
            The specified Python interpreter build cannot be found.
    """

    check_python_interpreter_not_found_error(command)

def check_python_interpreter_not_found_error(command):
    if 'sys.executable' in command:
        if command.startswith('python') or command.startswith('py'):
            version = get_python_version_from_command(command)
            raise PythonInterpreterNotFoundError(version)
        else:
            python_build = command.split()[0]
            raise PythonBuildNotFoundError(python_build)






import platform
import re

from packaging.version import Version
from pathlib import Path
from typing import NamedTuple

import wox.subprocessor as subprocessor

from wox.logger.logger_setup import wox_logger
from wox.utils import normalize_python_version
from wox.wox_exceptions import NotPythonVenvWarning


class PythonInterpreter(NamedTuple):
    """
        Parameters
        ----------
        version: pachaking.version.Version

        path: pathlib.Path
    """
    version: Version
    path: Path

def python_discovery(env, python_cmd = None):
    """
        Discovers and creates a Python interpreter with version and path.

        Parameters
        ----------
        env : str
            Name of the virtual environment
        python_cmd : str, default None
            Path to a Python interpreter provided by python_build from .toml file.

        Returns
        -------
        PythonInterpreter
            Python interpreter with its version and path.
    """

    if python_cmd is None:
        if env.startswith('py') and env[2] == '3':
            version = get_interpreter_version(env)
            wox_logger.debug(f'Looking for an interpreter version {version}.')

            match platform.system():
                case 'Windows':
                    python_cmd = 'py', f'-{version}'
                case 'Linux':
                    python_cmd = f'python{version}'
        else:
            wox_logger.debug('Looking for the system Python interpreter.')
            python_cmd = find_python_system()

    script = "import platform, sys; print(platform.python_version(), sys.executable)"

    python_information = subprocessor.run_command(
        f'{python_cmd} -c "{script}"',
        unsafe_command = True,
        return_output = True
    )

    version, path = python_information.split()

    wox_logger.debug(f"Interpreter found with version '{version}' on path '{path}'!")

    return PythonInterpreter(Version(version), Path(path))

def get_interpreter_version(env):
    """
        Extracts version from name of virtual environments following
        the nomenclature py<version> (e.g. py311).

        Parameters
        ----------
        env : str
            Name of the virtual environment.

        Returns
        -------
        str
            Python version using the major.minor scheme.
    """
    version_tag = env.lstrip('py')
    version = normalize_python_version(version_tag)

    return version

def find_python_system():
    """
        Finds the system Python interpreter.

        Returns
        -------
        str
            Command to start Python by default on the system.
    """
    match platform.system():
        case 'Windows':
            python_cmd = 'py'
        case 'Linux':
            python_cmd = 'python3'

    return python_cmd

def get_python_version_from_pyvenv_cfg(env_path):
    """
        Retrieves Python version of a virtual environment.

        Parameters
        ----------
        env_path : pathlib.Path
            Path to the virtual environment.

        Returns
        -------
        str
            Python version using the major.minor.fix scheme.
    """

    pyvenv_cfg = env_path / 'pyvenv.cfg'
    if pyvenv_cfg.exists():
        with open(pyvenv_cfg) as file:
            for line in file:
                if line.startswith('version'):
                    version = line.split('=')[1].strip()
                    version = re.findall(r'\d+', version)
                    return '.'.join(version[:3])
    else:
        raise NotPythonVenvWarning(env_path)

def locate_venv_python_exe(env_path):
    """
        Locates the Python interpreter of a virtual environment.

        Parameters
        ----------
        env_path : pathlib.Path
            Path to the virtual environment.

        Returns
        -------
        PythonInterpreter
            Python interpreter with its version and path.
    """

    version = get_python_version_from_pyvenv_cfg(env_path)

    python_paths = [
        env_path / 'Scripts' / 'python.exe',
        env_path / 'Scripts' / 'python',
        env_path / 'bin' / 'python',
        env_path / 'bin' / 'python3',
        env_path / 'bin' / 'python.exe',
        env_path / 'bin' / 'python3.exe',
    ]

    for path in python_paths:
        if path.exists():
            python_path = path

            wox_logger.debug(
                f"Interpreter located for venv '{env_path.name}' with version {version}"
                f" successfully at '{python_path}'!"
            )

            return PythonInterpreter(Version(version), python_path)



import os
import pytest

# Run tests as maintainer
# os.environ['RUN_MAINTAINER_TESTS'] = '1'

maintainer_only = pytest.mark.skipif(
    os.environ.get('RUN_MAINTAINER_TESTS') != '1',
    reason='Maintainer-only test, skipped because RUN_MAINTAINER_TESTS is not set'
)

pytest_plugins = [
    'tests.fixtures.cli',
    'tests.fixtures.paths',
    'tests.fixtures.temp_file_system'
]



import logging

from wox.logger.constants import (
    CONTEXTUAL_INFO,
    DATEFMT,
    STYLE
)


contextual_data = ':'.join(f'{{{key}}}' for key in CONTEXTUAL_INFO)
format = f'[{contextual_data}|{{levelname:<8}}({{levelno}})] {{message}}'

formatter = logging.Formatter(
    # https://docs.python.org/3/library/logging.html#logrecord-attributes
    fmt = format,
    datefmt = DATEFMT,
    style = STYLE,
    defaults = CONTEXTUAL_INFO
)



import logging
import sys

from wox.logger.handlers.stream_handler import initialize_console_handler

from wox.logger.handlers.log_file_handler import initialize_log_file_handler


console_handler = initialize_console_handler()
log_file_handler = initialize_log_file_handler()

wox_logger = logging.getLogger('wox')
wox_logger.setLevel(logging.INFO)
wox_logger.addHandler(console_handler)
if log_file_handler:
    wox_logger.addHandler(log_file_handler)
else:
    wox_logger.addHandler(logging.NullHandler())

if 'pytest' in sys.modules:
    wox_logger.propagate = True
else:
    wox_logger.propagate = False



# coverage run -m pytest -rA tests/test_configurator.py --no-header --html=./reports/pytest/report.html --self-contained-html && coverage html #noqa E501

import copy
import pytest

from wox.configurator import configuration_creation, session_setup
from wox.wox_exceptions import NoTaskError
from wox.wox_steward import WoxSteward


# python -B -m pytest -rA -k "test_configuration_creation"
def test_configuration_creation(monkeypatch, subtests, wox_toml_path):
    """
        Each environment configuration must be correctly built from
        wox.toml, applying its specific parameters when defined, or
        falling back to default settings otherwise.

        Environment-specific parameters must not propagate to other
        environments.
    """

    control_configs = {
        'testing': {
            'py36': {
                'deps': [
                    'pytest-7.0.1-py3-none-any.whl',
                    'tomli-1.2.3-py3-none-any.whl',
                ],
                'commands': ['python -B -m pytest'],
            },
            'default': {
                'deps': ['pytest'],
                'commands': ['python -B -m pytest'],
            },
        },
        'doc': {
            'default': {
                'deps': ['furo'],
                'commands': [
                    'python -m sphinx.cmd.quickstart docs --sep -q -p tuto -a wbarillon'
                ],
            }
        },
        'pixegami-hello': {
            'default': {
                'deps': ['pixegami-hello==0.3.0'],
                'commands': ['pixegami-hello'],
            }
        },
        'benchmark': {
            'py314': {
                'python_build': '~/python313/bin/python3'
            },
            'py313-dev': {
                'python_build': '~/python313-optim/bin/python3'
            },
            'py313-prod': {
                'python_build': '~/python313-prod/bin/python3'
            },
            'py313-control': {
                'python_build': '~/python313/bin/python3'
            }
        }
    }

    monkeypatch.setattr(WoxSteward, 'wox_toml', wox_toml_path)

    wox_data = WoxSteward().wox_data

    wox_configuration = wox_data.pop('wox_configuration') #noqa F841

    for task, settings in wox_data.items():

        envlist = wox_data[task].pop('envlist')

        for env in envlist:

            config = configuration_creation(env, settings, envlist)

            with subtests.test(task = task, env = env):
                assert config == control_configs[task][env], (
                    f"Configuration mismatch for task '{task}' and environment '{env}'.\
                    \n"
                )

class TestSessionSetup:
    """
        session_setup() must handle conflicts between wox_session_default, ``--task`` and
        ``exclude-task`` following requirements in Notes section of session_setup.
    """

    wox_configuration_with_wox_session_default = {
        'wox_configuration': {
            'recreate_envs': True,
            'wox_session_default': ['cowsay', 'testing']
        },
        'testing': {},
        'cowsay': {},
        'pixegami': {},
        'bash': {},
        'doc': {}
    }
    wox_configuration_without_wox_session_default = {
        'wox_configuration': {'recreate_envs': True},
        'testing': {},
        'cowsay': {},
        'pixegami': {},
        'bash': {},
        'doc': {}
    }

    test_session_setup_cases = [
        # wox
        (
            {'verbose': False, 'task': None, 'exclude_task': None},
            wox_configuration_without_wox_session_default,
            ['testing', 'cowsay', 'pixegami', 'bash', 'doc']
        ),
        # wox with wox_session_default
        (
            {'verbose': False, 'task': None, 'exclude_task': None},
            wox_configuration_with_wox_session_default,
            ['testing', 'cowsay']
        ),
        # wox --task doc without wox_session_default
        (
            {'verbose': False, 'task': ['doc'], 'exclude_task': None},
            wox_configuration_without_wox_session_default,
            ['doc']
        ),
        # wox --task doc with wox_session_default
        (
            {'verbose': False, 'task': ['doc'], 'exclude_task': None},
            wox_configuration_with_wox_session_default,
            ['doc']
        ),
        # wox --exclude-task cowsay bash without wox_session_default
        (
            {'verbose': False, 'task': None, 'exclude_task': ['cowsay', 'bash']},
            wox_configuration_without_wox_session_default,
            ['testing', 'pixegami', 'doc']
        ),
        # wox --exclude-task cowsay with wox_session_default
        (
            {'verbose': False, 'task': None, 'exclude_task': ['cowsay']},
            wox_configuration_with_wox_session_default,
            ['testing']
        ),
        # wox --task pixegami bash --exclude-task bash
        (
            {'verbose': False, 'task': ['pixegami', 'bash'], 'exclude_task': ['bash']},
            wox_configuration_with_wox_session_default,
            ['pixegami']
        )
    ]

    # python -B -m pytest -rA -k "test_session_setup"
    @pytest.mark.parametrize('cli_options, wox_data, expected', test_session_setup_cases)
    def test_session_setup(self, cli_options, wox_data, expected):

        wox_data = copy.deepcopy(wox_data)

        wox_configuration, wox_data = session_setup(cli_options, wox_data)

        assert list(wox_data) == expected

class TestNoTaskError:
    """
        session_setup() must raise NoTaskError when ``--task`` and ``--exclude-task``
        have the same tasks.
    """

    test_no_task_error_cases = [
        # wox --task pixegami --exclude-task pixegami
        (
            {'verbose': False, 'task': ['pixegami'], 'exclude_task': ['pixegami']},
            {'wox_configuration': {'recreate_envs': True},
                'testing': {},
                'cowsay': {},
                'pixegami': {},
                'bash': {},
                'doc': {}
            }
        ),
        # wox --task testing --exclude-task testing with wox_session_default
        (
            {'verbose': False, 'task': ['testing'], 'exclude_task': ['testing']},
            {'wox_configuration': {
                'recreate_envs': True,
                'wox_session_default': ['cowsay', 'testing']
                },
                'testing': {},
                'cowsay': {},
                'pixegami': {},
                'bash': {},
                'doc': {}
            }
        ),
        # wox --exclude-task cowsay testing with wox_session_default
        (
            {'verbose': False, 'task': None, 'exclude_task': ['cowsay', 'testing']},
            {'wox_configuration': {
                'recreate_envs': True,
                'wox_session_default': ['cowsay', 'testing']
                },
                'testing': {},
                'cowsay': {},
                'pixegami': {},
                'bash': {},
                'doc': {}
            }
        )
    ]

    # python -B -m pytest -rA -k "test_no_task_error"
    @pytest.mark.parametrize('cli_options, wox_data', test_no_task_error_cases)
    def test_no_task_error(self, cli_options, wox_data):

        wox_data = copy.deepcopy(wox_data)

        with pytest.raises(NoTaskError):
            wox_configuration, wox_data = session_setup(cli_options, wox_data)



# coverage run -m pytest -rA tests/test_utils.py --no-header --html=./reports/pytest/report.html --self-contained-html && coverage html #noqa E501

import pytest

from wox.subprocess_error_detection.utils import detect_sentence_with_fill_in_blank
from wox.utils import auto_clean_logs, get_python_version_from_command


# python -B -m pytest -rA -k "test_auto_clean_logs"
def test_auto_clean_logs(monkeypatch, tmp_path):
    """
        auto_clean_logs must delete the most recent file created.
    """

    logs_dir = tmp_path / '.logs'
    logs_dir.mkdir()

    log_1 = logs_dir / 'output-1.log'
    log_1.touch()

    log_2 = logs_dir / 'output-2.log'
    log_2.touch()

    before = list(logs_dir.glob('output*'))

    assert len(before) == 2

    monkeypatch.setattr('wox.utils.DOT_WOX', tmp_path)

    auto_clean_logs()

    after = list(logs_dir.glob('output*'))

    assert len(after) == 1

    assert after[0].name == 'output-1.log'

class TestGetPythonVersionFromCommand:
    """
        get_python_version_from_command() must return a Python version
        following major.minor scheme.
    """

    test_get_python_version_from_command_cases = [
        {
            'input': 'python3.6',
            'expected': '3.6'
        },
        {
            'input': 'py -3.6',
            'expected': '3.6'
        }
    ]

    # python -B -m pytest -rA -k "test_get_python_version_from_command"
    @pytest.mark.parametrize('test_case_', test_get_python_version_from_command_cases)
    def test_get_python_version_from_command(self, test_case_):

        version = get_python_version_from_command(test_case_['input'])

        assert version == test_case_['expected']

# python -B -m pytest -rA -k "test_detect_sentence_with_fill_in_blank"
@pytest.mark.parametrize(
    'text', [
        'The world is full of flowers ready to blossom',
        'The world is\n full of wonders\n ready to blossom'
    ]
)
def test_detect_sentence_with_fill_in_blank(text):
    """
        detect_sentence_with_fill_in_blank() must return True if the sentence
        with fill-in-the-blanks is found in the text.
    """

    assert detect_sentence_with_fill_in_blank(text, 'full of * ready')



import argparse
import sys

from wox._version import __version__


def parser_definition():
    parser = argparse.ArgumentParser(
        prog = 'wox',
        description = (
            'A declarative CLI that runs reproducible tasks inside '
            'isolated Python environments'
        ),
        formatter_class = argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument(
        '-V', '--version',
        action = 'version',
        version = f'%(prog)s {__version__}',
        help = 'prints the version of wox and exit'
    )
    parser.add_argument(
        '-v', '--verbose',
        action = 'store_true',
        help = 'enables verbose output'
    )
    parser.add_argument(
        '--task',
        nargs = '+',
        help = 'selects the tasks to execute among those in the .toml file'
    )
    parser.add_argument(
        '--exclude-task',
        nargs = '+',
        help = 'excludes the tasks to execute among those in the .toml file'
    )
    parser.add_argument(
        '--purge',
        action = 'store_true',
        help = (
            'deletes all directories and files in .wox except .logs, .gitignore '
            'and CACHEDIR.TAG'
        )
    )

    return parser.parse_args()

def main_cli():
    """
        Wox entrypoint.

        Raises
        ------
        Exception
            Any Exception that is not a WoxException, that is,
            any exception not handled by wox.

        Notes
        -----
        Raised exception always return 1 as exit code, same value
        as ExitCode.RUNTIME_ERROR.
    """
    parser = parser_definition()
    cli_options = vars(parser)

    try:
        from wox.cli_control import cli_controller

        exit_code = cli_controller(cli_options)
        sys.exit(exit_code)
    except Exception:
        raise



# coverage run -m pytest -rA tests/test_wox_steward.py --no-header --html=./reports/pytest/report.html --self-contained-html && coverage html #noqa E501

import pytest
import tomllib

from wox.wox_exceptions import ConfigFileNotFoundError, MissingWoxSectionError
from wox.wox_steward import WoxSteward


# python -B -m pytest -rA -k "test_pyproject_toml_integrity"
def test_pyproject_toml_integrity(monkeypatch, pyproject_toml_path):
    """
        Extracting wox data from pyproject.toml must not
        alter the original file content.

        A control reference is loaded first, before wox_data
        process and the file is loaded again.

        Both versions must be identical.
    """

    with open(pyproject_toml_path, 'rb') as toml_file:
        control_loaded_pyproject_toml = tomllib.load(toml_file)

    monkeypatch.setattr(WoxSteward, 'wox_toml', pyproject_toml_path)

    WoxSteward().wox_data

    with open(pyproject_toml_path, 'rb') as toml_file:
        tested_loaded_pyproject_toml = tomllib.load(toml_file)

    assert control_loaded_pyproject_toml == tested_loaded_pyproject_toml

# python -B -m pytest -rA -k "test_load_wox_toml"
def test_load_wox_toml(monkeypatch, wox_toml_path):
    """
        wox_data must correctly extract Wox information from wox.toml.
    """

    with open(wox_toml_path, 'rb') as toml_file:
        loaded_wox_toml = tomllib.load(toml_file)

    monkeypatch.setattr(WoxSteward, 'wox_toml', wox_toml_path)

    wox_data = WoxSteward().wox_data

    assert loaded_wox_toml == wox_data

# python -B -m pytest -rA -k "test_wox_pyproject_equals_wox_toml"
def test_wox_pyproject_equals_wox_toml(monkeypatch, pyproject_toml_path, wox_toml_path):
    """
        Wox information extracted from pyproject.toml and wox.toml must yield
        identical data structures.
    """

    monkeypatch.setattr(WoxSteward, 'wox_toml', pyproject_toml_path)

    wox_data_from_pyproject = WoxSteward().wox_data

    monkeypatch.setattr(WoxSteward, 'wox_toml', wox_toml_path)

    wox_data_from_wox = WoxSteward().wox_data

    assert wox_data_from_pyproject == wox_data_from_wox

# python -B -m pytest -rA -k "test_config_file_not_found_error"
def test_config_file_not_found_error(cwd_tmp_path):
    """
        A workspace without wox.toml or pyproject.toml
        must throw ConfigFileNotFoundError.
    """

    with pytest.raises(ConfigFileNotFoundError):
        WoxSteward().wox_data

# python -B -m pytest -rA -k "test_missing_wox_section_error"
def test_missing_wox_section_error(tmp_path, cwd_tmp_path):
    """
        A workspace without wox.toml but containing pyproject.toml without
        a [wox] section must raise MissingWoxSectionError.
    """

    empty_pyproject_toml = tmp_path / 'pyproject.toml'
    empty_pyproject_toml.touch(exist_ok = True)

    with pytest.raises(MissingWoxSectionError):
        WoxSteward().wox_data

# python -B -m pytest -rA -k "test_dot_wox_setup"
def test_dot_wox_setup(monkeypatch, tmp_path, cwd_tmp_path):
    """
        Upon setup, .wox must contain both .gitignore and CACHEDIR.TAG files.
    """

    control_gitignore_content = (
        '# Automatically created by wox.\n'
        '*\n'
    )

    control_cachedir_tag = (
        'Signature: 8a477f597d28d172789f06886806bc55\n'
        '# This directory is managed by Wox.\n'
        '# It contains cache data such as logs and virtual environments.\n'
        '# See: https://bford.info/cachedir/\n'
    )

    monkeypatch.setattr('wox.wox_steward.DOT_WOX', tmp_path)

    WoxSteward().dot_wox_setup()

    gitignore_path = tmp_path / ".gitignore"
    cachedir_tag_path = tmp_path / "CACHEDIR.TAG"

    assert gitignore_path.exists()
    assert cachedir_tag_path.exists()

    assert gitignore_path.read_text() == control_gitignore_content
    assert cachedir_tag_path.read_text() == control_cachedir_tag



from packaging.specifiers import SpecifierSet

import wox.subprocessor as subprocessor

from wox.constants import DOT_WOX, MSV_VIRTUALENV
from wox.logger.logger_setup import wox_logger
from wox.wox_exceptions import MissingWoxKeywordError
from wox.utils import rmtree_with_retry


def venv_deletion(tasks):
    """
        Deletes virtual environments used during the wox session which
        were created in a previous wox session.

        Parameters
        ----------
        tasks : dict
            Tasks declared in the .toml file.

        Raises
        ------
        MissingWoxKeywordError
            Required wox keyword missing in .toml file.
    """

    for task in tasks:

        try:
            envlist = tasks[task]['envlist']
        except KeyError as e:
            raise MissingWoxKeywordError(e)

        for env in envlist:

            env_path = DOT_WOX / env
            if env_path.exists():
                wox_logger.info(f'Deletion of venv {env}.')
                try:
                    rmtree_with_retry(env_path)
                except Exception as e:
                    wox_logger.warning(f'{type(e).__name__}: {e}')
                    continue
                else:
                    wox_logger.debug(f'Venv {env} deleted!')
            else:
                wox_logger.warning(
                    f"Unable to delete '{env}': virtual environment "
                    f"not found on '{env_path}'"
                )

def create_venv(env_path, python_interpreter, with_uv = False):
    """
        Creates virtual environment running during the wox session.

        Parameters
        ----------
        env_path : pathlib.Path
            Path to the virtual environment.
        python_interpreter : PythonInterpreter
            Discovered Python interpreter with its version and path.
        with_uv : bool
            If True, switches pip to uv.

        Raises
        ------
        PythonInterpreterUnstableError
            The Python interpreter crashes at the C-level.
    """

    wox_logger.debug(f'Creation of venv {env_path.name} on path {env_path}.')

    if with_uv and python_interpreter.version in SpecifierSet(MSV_VIRTUALENV):
        subprocessor.run_command(
            f'uv venv --python {python_interpreter.path} {str(env_path)}',
            python_interpreter = python_interpreter
        )
    else:
        subprocessor.run_command(
            f'{python_interpreter.path} -m venv {str(env_path)}',
            python_interpreter = python_interpreter
        )

    wox_logger.debug(f'Venv {env_path.name} created!')





import os
import platform
import shlex
import shutil
import subprocess

from io import StringIO

import mslex

from wox.logger.filters import WoxContext
from wox.logger.logger_setup import wox_logger
from wox.wox_exceptions import (
    CommandNotFoundError,
    PythonInterpreterCallError,
    WoxException
)


# Only used by pytest summoning subprocessor directly without passing by main entrypoint.
wox_logger.addFilter(WoxContext({'task': 'pytest', 'env': 'subprocess'}))

def sanitize_command(command, python_interpreter_path = None):
    """
        Sanitizes a command before execution by resolving its executable path.

        Parameters
        ----------
        command : str
            The command line string to execute.
        python_interpreter_path : pathlib.Path or None
            Path to the Python binary of a virtual environment.
            If None, uses the system PATH.

        Raises
        ------
        PythonInterpreterCallError
            Command tries to launch an interactive Python session.
        CommandNotFoundError
            Executable for the command not found.

        Returns
        -------
        list[str]
            A list representing the sanitized command ready for subprocess.
    """

    match platform.system():
        case 'Linux':
            command = shlex.split(command)
        case 'Windows':
            command = mslex.split(command, like_cmd = False)

    # --- Special case: user launched "python" alone ---
    if (command[0] == 'py' or command[0] == 'python') and len(command) == 1:
        raise PythonInterpreterCallError(command[0])

    cmd, options = command[0], command[1:]

    # --- PATH resolution ---
    if python_interpreter_path:
        # Add venv's bin folder first in PATH
        path = str(python_interpreter_path.parent) + os.pathsep + os.environ['PATH']
        # Replace `python` by explicit path to the venv interpreter
        if cmd == 'python':
            cmd = str(python_interpreter_path)
    else:
        path = os.environ['PATH']

    # --- Resolve executable ---
    executable = shutil.which(cmd = cmd, path = path)
    if not executable:
        raise CommandNotFoundError(cmd)

    return [executable] + options

def run_command(command,
        unsafe_command = False, shell = False,
        return_output = False,
        python_interpreter = None,
        timeout = 2
    ):
    """
        Executes the command in subprocess.

        Parameters
        ----------
        command : str
            Command passed to subprocess.
        unsafe_command : bool, default False
            If True, the command is not sanitized and use subprocess as shell.
        shell : bool, default False
            Use subprocess as shell, which implies security issues.
        return_output : bool, default False
            If True, the subprocess returns the output from its stderr.
        PythonInterpreter
            Python interpreter with its version and path.
        timeout : int, default 10
            Delay in seconds before subprocess throws a TimeoutError.

        Returns
        -------
        str
            The output of subprocess' stderr. If return_output is True.

        Raises
        ------
        DependencyConflictError
            Dependency conflict occurs during package installation.
        DiskFullError
            Disk is full and command cannot write data.
        PipTimeoutError
            Pip cannot reach package server.
        PackageNotFoundError
            Package is missing or wrongly spelled.
        PermissionDeniedError
            Permissions prevent writing data.
        PythonInterpreterNotFoundError
            The specified Python interpreter cannot be found.
        PythonInterpreterUnstableError
            The Python interpreter crashes at the C-level.
        subprocess.CalledProcessError
            Fallback for unhandled errors.
    """

    if unsafe_command:
        shell = True
    else:
        command = sanitize_command(
            command, python_interpreter.path if python_interpreter else None
        )

    env = os.environ.copy()
    if python_interpreter is not None:
        env['PATH'] = str(python_interpreter.path.parent) + os.pathsep + os.environ['PATH']
    if any(part.endswith('pytest') or part == 'pytest' for part in command):
        env['WOX_SUBPROCESSOR_NO_LOG'] = '1'

    wox_logger.info(f"Subprocess launched with command '{command}'.")
    wox_logger.info('Subprocess output - start')
    wox_logger.info('-'*25)

    with StringIO() as buffer:
        with subprocess.Popen(
            args = command,
            stdout = subprocess.PIPE,
            stderr = subprocess.PIPE,
            shell = shell,
            env = env,
            text = True,
            encoding = 'utf-8',
            errors = 'replace'
        ) as process:

            for line in process.stdout:
                wox_logger.info(line.rstrip('\r\n'))
                if return_output:
                    buffer.write(line)

            for line in process.stderr:
                wox_logger.info(line.rstrip('\r\n'))
                buffer.write(line)

            process.wait(timeout = timeout)

            wox_logger.info('-'*23)
            wox_logger.info('Subprocess output - end')

            if process.returncode != 0:

                stderr_output = buffer.getvalue()
                if not stderr_output:
                    subprocess_error = subprocess.CalledProcessError(
                        process.returncode, command
                    )
                    stderr_output = str(subprocess_error).lower()

                try:
                    if unsafe_command is True and 'python' in command:
                        import wox.subprocess_error_detection.naive_python as error_detector

                        error_detector.detect_python_discovery_errors(command)
                    elif (
                        (len(command) > 3 and command[3] == 'pip') or \
                        (len(command) > 2 and command[2] == 'venv')
                    ):
                        import wox.subprocess_error_detection.pip_install as error_detector

                        error_detector.detect_pip_errors(
                            stderr_output,
                            command,
                            python_interpreter if python_interpreter else None
                        )

                    elif os.path.basename(command[0]) == 'uv':
                        import wox.subprocess_error_detection.uv_install as error_detector

                        error_detector.detect_uv_errors(
                            stderr_output,
                            command,
                            python_interpreter
                        )
                except WoxException:
                    raise
                else:
                    raise subprocess.CalledProcessError(process.returncode, command)

            if return_output:
                return buffer.getvalue().strip()



import os
import platform
import pytest
import shutil
import time

from pathlib import Path

import wox.subprocessor as subprocessor

from tests.resources.constants import RAMDISK_PATH


@pytest.fixture(scope = 'session')
def tmpfs():
    """
        Mounts a RAM disk and fills it with data.

        Notes
        -----
        During tests, `test_pip_disk_full()` can get stuck on the pip install step.
        To work around this, the test uses a timeout and retries. On retry, the test
        manages to attempt the installation on the RAM disk and raises DiskFullError.

        If the fixture had function scope, teardown would trigger between retries
        and the retry would create a new RAM disk, breaking the test.

        Setting the scope to 'session' ensures the fixture persists across retries.

        Additionally, this fixture is shared between the pip install test and
        its counterpart, uv pip install.
        Somehow, if pip has already attempted to install pandas in the RAM disk,
        the uv test fails.

        The proper workaround is to run the uv test first, then the pip test.

        By doing so, the pip test doesn’t even need to retry, since the uv test
        has already prepared the fixture.
    """

    if platform.system() != 'Linux':
        pytest.skip('tmpfs only works on Linux')

    if not shutil.which('sudo'):
        pytest.skip('tmpfs requires sudo')

    if not RAMDISK_PATH.exists():
        Path.mkdir(RAMDISK_PATH, parents = True)

        filler_file = RAMDISK_PATH / 'fill.txt'

        subprocessor.run_command(f'sudo mount -t tmpfs -o size=1M tmpfs {RAMDISK_PATH}')

        with open(f'{filler_file}', 'wb') as file:
            file.write(os.urandom(1024 * 1024))

    yield

    subprocessor.run_command(f'sudo umount {RAMDISK_PATH}')
    time.sleep(1)
    Path.rmdir(RAMDISK_PATH)



import pytest


@pytest.fixture
def patched_wox_command(monkeypatch):
    def patched_parser_definition():
        """
            Emulates a 'wox' command executed by user.
        """

        class Args:
            task = None
            version = None
        return Args()

    return monkeypatch.setattr('wox.cli.parser_definition', patched_parser_definition)



import ast
import multiprocessing
import re


def run_with_timeout_process(func, timeout, *args, **kwargs):
    """
        Executes a function in a separate process and terminates it if
        the timeout is reached.

        Parameters
        ----------
        func : callable
            The function to execute in the subprocess.
        timeout : int
            Maximum allowed execution time in seconds.

        Raises
        ------
        Exception
            Any exception raised by the function.
        TimeoutError
            If the function execution exceeds the timeout.
    """

    queue = multiprocessing.Queue()

    def target():
        try:
            func(*args, **kwargs)
            queue.put(None)
        except Exception as e:
            queue.put(e)

    proc = multiprocessing.Process(target = target)
    proc.start()
    proc.join(timeout)
    if proc.is_alive():
        proc.terminate()
        proc.join()
        raise TimeoutError(f'Timeout de {timeout}s atteint')

    exception = queue.get()
    if exception:
        raise exception

def extract_traceback_from_caplog(caplog):
    """
        Extracts the traceback section from a pytest caplog output.

        Parameters
        ----------
        caplog : str
            Raw text content captured from pytest's caplog.

        Returns
        -------
        str
            The traceback text, formatted exactly as it would appear
            in the terminal output.
    """

    pattern = re.compile(
        r"Subprocess output - start\n"
        r"(.*?)"
        r"Subprocess output - end",
        re.DOTALL
    )
    matches = pattern.findall(caplog)
    if not matches:
        return None

    last_block = matches[-1].strip()

    # Supprime les préfixes de log de type 'INFO wox:subprocessor.py:198 '
    last_block = re.sub(
        r"^\w+\s+\S+:\d+\s", "",
        last_block,
        flags=re.MULTILINE
    )

    # Supprime les lignes de séparation (----)
    last_block = re.sub(
        r"^-{3,}\s*$", "",
        last_block,
        flags=re.MULTILINE
    )

    # Supprime toute ligne qui commence par un préfixe de log résiduel
    last_block = re.sub(
        r"^(DEBUG|INFO|WARNING|ERROR|CRITICAL)\s+.*$", "",
        last_block,
        flags=re.MULTILINE
    )

    return last_block.strip()

def extract_dict_from_caplog(caplog, marker):
    """
        Extracts a dict from a pytest caplog output.

        Parameters
        ----------
        caplog : str
            Raw text content captured from pytest's caplog.
        marker: str
            String helping identifying the log storing the dict.

        Returns
        -------
        dict
            The dict recreated from its string form from log.

        Notes
        -----
        Extracting a dict from logs implies to turn the string representing the dict
        into an actual dict.

        ast is a library that manipulates Python code as data structure.
        Among its features, literal_eval() turns any string into an actual object, which
        is perfect for this use case.

        https://docs.python.org/3/library/ast.html#ast.literal_eval
    """

    log_record = [
        record for record in caplog.records \
        if marker in record.message
    ][0].msg

    match = re.search(r'(\{.*\})', log_record, re.DOTALL)
    if match:
        return ast.literal_eval(match.group(1))



"""
    This module contains all error that may occur during the management of
    virtual environments handled by uv.
"""

import re

from wox.subprocess_error_detection.utils import (
    detect_error_from_stderr,
    detect_sentence_with_fill_in_blank
)
from wox.utils import get_python_version_from_command
from wox.wox_exceptions import (
    DependencyConflictError,
    DiskFullError,
    PackageNotFoundError,
    PermissionDeniedError,
    PipTimeoutError,
    PythonInterpreterUnstableError
)


def detect_uv_errors(stderr_output, command, python_interpreter):
    """
        Detects uv-related errors upon dependencies installation.

        Parameters
        ----------
        stderr_output : str
            Output captured from the subprocess stderr stream.
        command : str
            Command executed by subprocess.
        python_interpreter : PythonInterpreter
            Python interpreter with its version and path.

        Raises
        ------
        DependencyConflictError
            Dependency conflict occurs during package installation.
        DiskFullError
            Disk is full and command cannot write data.
        PipTimeoutError
            Pip cannot reach package server.
        PackageNotFoundError
            Package is missing or wrongly spelled.
        PermissionDeniedError
            Permissions prevent writing data.
        PythonInterpreterUnstableError
            The Python interpreter crashes at the C-level.
    """

    check_dependency_conflict_error(stderr_output, command)
    check_disk_full_error(stderr_output, command)
    check_pip_timeout_error(stderr_output, command)
    check_package_not_found_error(stderr_output, command, python_interpreter.path)
    check_permission_denied_error(stderr_output, command)
    check_python_interpreter_unstable_error(stderr_output,
        command,
        python_interpreter.version
    )

def check_dependency_conflict_error(stderr_output, command):

    if (
        'install' in command and \
        detect_sentence_with_fill_in_blank(stderr_output,
            'your requirements and * are incompatible'
        )
    ):
        raise DependencyConflictError

def check_disk_full_error(stderr_output, command):
    if 'install' in command and detect_error_from_stderr(stderr_output, ['error 28']):
        raise DiskFullError

def check_pip_timeout_error(stderr_output, command):
    if 'install' in command and detect_error_from_stderr(stderr_output,
        ['connection refused (os error 111)', 'error sending request for url']
    ):
        raise PipTimeoutError

def check_package_not_found_error(stderr_output, command, python_interpreter_path):
    if ('install' in command and \
        detect_error_from_stderr(stderr_output, ['there are no versions'])
    ):
        environment = python_interpreter_path.parent.parent.name
        match = re.search(r'there are no versions of (\S+)', stderr_output)
        if match:
            missing_package = match.group(1)
        raise PackageNotFoundError(missing_package, environment)

def check_permission_denied_error(stderr_output, command):
    if (
        'install' in command and \
        detect_error_from_stderr(stderr_output,['os error 13'])
    ):
        raise PermissionDeniedError

def check_python_interpreter_unstable_error(stderr_output, command, version):
    if (
        (
            ('venv' in command or 'install' in command) and \
            detect_error_from_stderr(stderr_output,
                [
                    'died with', 'sigsegv',
                    'segmentation fault', 'core dumped',
                    'problem confirming the ssl certificate',
                    'ssl module is not available',
                    'ssl module in python is not available'
                ]
            )
        ) or (
            'cannot open shared object file' in stderr_output.lower() and \
            'libpython' in stderr_output.lower()
        )
    ):
        if not version:
            version = get_python_version_from_command(command)
        raise PythonInterpreterUnstableError(version)



# coverage run -m pytest -rA tests/test_venv_management.py --no-header --html=./reports/pytest/report.html --self-contained-html && coverage html #noqa E501

import wox.python_discoverer as python_discoverer
import wox.venv_management as venv_manager


# python -B -m pytest -rA -k "test_venv_management"
def test_venv_management(tmp_path, monkeypatch):
    """
        Ensures the virtual environment management workflow functions as expected.

        This nominal test validates the successful execution of the following steps:

        - creation of a virtual environment;
        - location of the Python interpreter within that environment;
        - deletion of the environment afterward.
    """

    env = 'default'

    env_path = tmp_path / env

    python_interpreter = python_discoverer.python_discovery(env)

    venv_manager.create_venv(env_path, python_interpreter)

    venv_python_interpreter = python_discoverer.locate_venv_python_exe(env_path)

    assert venv_python_interpreter.path.exists()

    monkeypatch.setattr('wox.venv_management.DOT_WOX', tmp_path)

    wox_data = {
        'testing': {'envlist': [env]}
    }

    venv_manager.venv_deletion(wox_data)

    assert not venv_python_interpreter.path.exists()

# python -B -m pytest -rA -k "test_venv_deletion_not_found"
def test_venv_deletion_not_found(monkeypatch, tmp_path, caplog):
    """
        Ensures graceful handling when attempting to delete a non-existent
        virtual environment.

        venv_deletion() must log a specific warning message if the target
        environment cannot be found instead of raising an exception.
    """

    env = 'default'

    wox_data = {
        'testing': {'envlist': [env]}
    }

    control_log_record_message = (
        f"Unable to delete '{env}': virtual environment "
        f"not found on '{tmp_path / env}'"
    )

    monkeypatch.setattr('wox.venv_management.DOT_WOX', tmp_path)

    venv_manager.venv_deletion(wox_data)

    assert len(caplog.records) == 1

    log_record_message = caplog.records[0].msg

    assert control_log_record_message == log_record_message

# python -B -m pytest -rA -k "test_venv_deletion_fail_rmtree"
def test_venv_deletion_fail_rmtree(monkeypatch, tmp_path, caplog):
    """
        Ensures graceful handling when shutil.rmtree() fails
        to delete a directory tree.

        venv_deletion() must log the exception message instead
        of propagating the raised exception.
    """

    def patched_rmtree(path):
        raise PermissionError("Access denied (patched)")

    env = 'default'

    env_path = tmp_path / env

    wox_data = {
        'testing': {'envlist': [env]}
    }

    control_log_record_message = 'PermissionError: Access denied (patched)'

    python_interpreter = python_discoverer.python_discovery(env)

    venv_manager.create_venv(env_path, python_interpreter)

    venv_python_interpreter = python_discoverer.locate_venv_python_exe(env_path)

    assert venv_python_interpreter.path.exists()

    monkeypatch.setattr('shutil.rmtree', patched_rmtree)
    monkeypatch.setattr('wox.venv_management.DOT_WOX', tmp_path)

    venv_manager.venv_deletion(wox_data)

    log_record_message = caplog.records[-1].msg

    assert control_log_record_message == log_record_message



import time

from packaging.specifiers import SpecifierSet

import wox.configurator as configurator
import wox.python_discoverer as python_discoverer
import wox.subprocessor as subprocessor
import wox.wox_summary as wox_summary
import wox.venv_management as venv_manager

from wox.constants import (
    DOT_WOX,
    MSV_UV,
    ExitCode
)
from wox.logger.filters import WoxContext
from wox.logger.logger_setup import wox_logger
from wox.wox_exceptions import MissingWoxKeywordError, NotPythonVenvWarning


def main(cli_options, wox_data):
    """
        Entrypoint for a wox session.

        Parameters
        ----------
        cli_options : dict
            Options provided on the command line.
        wox_data : dict
            Contains informations provided by the .toml file.

        Returns
        -------
        exit_code : int

        Raises
        ------
        Exception
            Any exceptions caught during the session.
    """

    start_time = time.perf_counter()
    exit_code = ExitCode.SUCCESS.value
    wox_session_summary = []

    logger_extra_fields = WoxContext()
    wox_logger.addFilter(logger_extra_fields)
    if cli_options.get('verbose', False):
        wox_logger.setLevel('DEBUG')

    wox_logger.debug(f'wox launched with options {cli_options}')

    try:
        wox_configuration, tasks = configurator.session_setup(cli_options, wox_data)
    except Exception:
        raise

    with_uv = wox_configuration.get('with_uv', False)

    wox_logger.debug('Wox session initialized with:')
    wox_logger.debug(f'- wox configuration {wox_configuration} ;')
    wox_logger.debug(f'- task settings {tasks} ;')

    if wox_configuration.get('recreate_envs', None):
        venv_manager.venv_deletion(tasks)

    for task, settings in tasks.items():

        try:
            envlist = tasks[task].pop('envlist')
        except KeyError as e:
            raise MissingWoxKeywordError(e)

        for env in envlist:

            logger_extra_fields.update_context({'task': task, 'env': env})

            execution_unit = {'execution_unit': f'{task}:{env}'}

            config = configurator.configuration_creation(env, settings, envlist)

            env_path = DOT_WOX / env

            if not env_path.exists():
                if (python_cmd := config.get('python_build')):
                    python_cmd = config['python_build']

                    wox_logger.debug(f"Use of specified python build '{python_cmd}'.")

                try:
                    python_interpreter = python_discoverer.python_discovery(
                        env, python_cmd
                    )
                except Exception as e:

                    wox_summary.update_execution_unit_failure(
                        execution_unit,
                        'Python discovery',
                        e
                    )
                    wox_session_summary.append(execution_unit)

                    wox_logger.error(e)
                    exit_code = ExitCode.RUNTIME_ERROR.value
                    continue

                try:
                    venv_manager.create_venv(env_path, python_interpreter, with_uv)
                except Exception as e:

                    wox_summary.update_execution_unit_failure(
                        execution_unit,
                        'Venv creation',
                        e
                    )

                    wox_session_summary.append(execution_unit)

                    wox_logger.error(e)
                    exit_code = ExitCode.RUNTIME_ERROR.value
                    continue

            try:
                venv_python_interpreter = python_discoverer.locate_venv_python_exe(env_path)
            except NotPythonVenvWarning as e:
                venv_python_interpreter = None
                wox_logger.warning(e)

            deps = ' '.join(config.get('deps', []))
            if deps:
                try:
                    if with_uv and venv_python_interpreter.version in SpecifierSet(MSV_UV):
                        subprocessor.run_command(
                            (f"uv pip install --python {venv_python_interpreter.path} "
                            f"{'-qqq' if not cli_options.get('verbose', False) else ''} "
                            f"{deps}"),
                            python_interpreter = venv_python_interpreter
                        )
                    else:
                        subprocessor.run_command(
                            (f"{venv_python_interpreter.path} -I -m pip install "
                            f"{'-qqq' if not cli_options.get('verbose', False) else ''} "
                            f"-U pip"),
                            python_interpreter = venv_python_interpreter
                        )
                        subprocessor.run_command(
                            (f"{venv_python_interpreter.path} -I -m pip install "
                            f"{'-qqq' if not cli_options.get('verbose', False) else ''} "
                            f"{deps}"),
                            python_interpreter = venv_python_interpreter
                        )
                except Exception as e:

                    wox_summary.update_execution_unit_failure(
                        execution_unit,
                        'Dependencies installation',
                        e
                    )
                    wox_session_summary.append(execution_unit)

                    wox_logger.error(e)
                    exit_code = ExitCode.RUNTIME_ERROR.value
                    continue

            for command in config.get('commands', []):
                try:
                    subprocessor.run_command(
                        command,
                        python_interpreter = venv_python_interpreter
                    )
                except Exception as e:

                    wox_summary.update_execution_unit_failure(
                        execution_unit,
                        'Commands execution',
                        e
                    )
                    wox_session_summary.append(execution_unit)

                    wox_logger.error(e)
                    exit_code = ExitCode.RUNTIME_ERROR.value
                    continue

            wox_summary.update_execution_unit_success(execution_unit)

            wox_session_summary.append(execution_unit)

    duration = round((time.perf_counter() - start_time), 3)
    wox_summary.render(wox_session_summary, duration)

    return exit_code



import colorlog

from wox.logger.constants import (
    CONTEXTUAL_INFO,
    DATEFMT,
    STYLE
)


contextual_data = ':'.join(f'{{{key}}}' for key in CONTEXTUAL_INFO)
format = (
    f'[{contextual_data}|{{levelname_log_color}}{{levelname:<8}}({{levelno}}){{reset}}] '
    f'{{levelname_log_color}}{{message}}{{reset}}'
)

colored_formatter = colorlog.ColoredFormatter(
    fmt = format,
    datefmt = DATEFMT,
    # https://pypi.org/project/colorlog/#description
    # The available colours are black, red, green, yellow, blue, purple, cyan and white.
    # See https://en.wikipedia.org/wiki/ANSI_escape_code.
    secondary_log_colors = {
        'levelname': {
            'DEBUG': 'cyan',
            'INFO': 'white',
            'WARNING': 'yellow',
            'ERROR': 'red',
            'CRITICAL': 'red,bold',
        }
    },
    style = STYLE,
    defaults = CONTEXTUAL_INFO
)



import logging
import os
import sys

from datetime import datetime

from wox.constants import DOT_WOX
from wox.logger.formatters.formatter import formatter


def initialize_log_file_handler():
    if os.getenv('WOX_SUBPROCESSOR_NO_LOG') == '1':
        return None
    if 'pytest' in sys.modules:
        log_path = DOT_WOX / '.logs' / 'tests'
        log_path.mkdir(parents = True, exist_ok = True)
        filename = log_path / f'output-{datetime.now():%Y-%m-%d_%H-%M-%S}.log'
    else:
        log_path = DOT_WOX / '.logs' / 'wox-session'
        log_path.mkdir(parents = True, exist_ok = True)
        filename = log_path/ f'output-{datetime.now():%Y-%m-%d_%H-%M-%S}.log'

    log_file_handler = logging.FileHandler(
        filename = filename,
        encoding = 'utf-8'
    )
    log_file_handler.setFormatter(formatter)

    return log_file_handler



# coverage run -m pytest -rA tests/test_main.py --no-header --html=./reports/pytest/report.html --self-contained-html && coverage html #noqa E501

"""
    This module tests wox with .toml files, ensuring the expected interactions
    between the toml file and the codebase.
"""

import pytest

from functools import cached_property
from pathlib import Path

from tests.conftest import maintainer_only
from tests.resources.utils import extract_traceback_from_caplog

from wox.cli import main_cli
from wox.cli_control import cli_controller
from wox.constants import ExitCode
from wox.wox_exceptions import WoxException
from wox.wox_steward import WoxSteward

from tests.resources.utils import extract_dict_from_caplog


# python -B -m pytest -rA -k "test_nominal_run"
def test_nominal_run(monkeypatch, tmp_path, patched_wox_command):
    """
        Wox session must return a SUCCESS exit code.
    """

    wox_toml_path = Path('tests/resources/config-files/nominal-run/wox.toml')

    monkeypatch.setattr(WoxSteward, 'wox_toml', wox_toml_path)
    monkeypatch.setattr('wox.main.DOT_WOX', tmp_path / '.wox')

    with pytest.raises(SystemExit) as e:
        main_cli()

    assert e.value.code == ExitCode.SUCCESS.value

# python -B -m pytest -rA -k "test_nominal_run_with_uv"
def test_nominal_run_with_uv(monkeypatch, tmp_path, patched_wox_command):
    """
        Wox session must return a SUCCESS exit code.
    """

    wox_toml_path = Path('tests/resources/config-files/nominal-run-with-uv/wox.toml')

    monkeypatch.setattr(WoxSteward, 'wox_toml', wox_toml_path)
    monkeypatch.setattr('wox.main.DOT_WOX', tmp_path / '.wox')

    with pytest.raises(SystemExit) as e:
        main_cli()

    assert e.value.code == ExitCode.SUCCESS.value

# python -B -m pytest -rA -k "test_unhandled_exception_caught"
def test_unhandled_exception_caught(monkeypatch, tmp_path, patched_wox_command):
    """
        Wox session must return an Exception that is not a WoxException.

        Notes
        -----
        In pytest, exceptions raised with pytest.raises are wrapped in
        an ExceptionInfo object. The actual exception instance can be
        accessed via `exception.value`.
    """

    wox_toml_path = Path('tests/resources/config-files/nominal-run/wox.toml')

    def patched_main(*args, **kwargs):
        raise ValueError("Boom")

    monkeypatch.setattr(WoxSteward, 'wox_toml', wox_toml_path)
    monkeypatch.setattr('wox.main.DOT_WOX', tmp_path / '.wox')
    monkeypatch.setattr('wox.cli_control.main', patched_main)

    with pytest.raises(Exception) as e:
        main_cli()

    assert isinstance(e.value, ValueError)
    assert not isinstance(e.value, WoxException)

# python -B -m pytest -rA -k "test_no_config_file"
def test_no_config_file(cwd_tmp_path):
    """
        Wox session must return a CONFIG_ERROR exit code due to
        missing config file.
    """

    exit_code = cli_controller({})

    assert exit_code == ExitCode.CONFIG_ERROR.value

class TestMissingWoxKeywordError:
    """
        Wox session must return a CONFIG_ERROR exit code due to
        MissingWoxKeywordError.

        The error must be logged.

        Notes
        -----
        `wox_data` is a `cached_property`, which is a descriptor. Descriptors are
        normally bound to the class at class creation time, when Python automatically
        calls their `__set_name__` method.

        When monkeypatching such an attribute after class creation, it is necessary
        to:
        1. replace the entire `cached_property` descriptor, not just its returned
        value;
        2. manually call `__set_name__` on the new descriptor so that it behaves
        correctly (i.e. knows its owning class and attribute name).

        Patching only the return value is insufficient, because a `cached_property`
        stores its computed result in the instance’s `__dict__`, and this cached
        value may already exist.
    """

    wrong_wox_data = [
        # typo in wox_configuration
        {
            'wox_configuratin': {'recreate_envs': True},
            'showcase': {
                'envlist': ['default'],
                'deps': ['cowsay'],
                'commands': ['cowsay -t "Hello Wox!" -c "tux"']
            }
        },
        # typo in envlist called in venv_deletion()
        {
            'wox_configuration': {'recreate_envs': True},
            'showcase': {
                'envlst': ['default'],
                'deps': ['cowsay'],
                'commands': ['cowsay -t "Hello Wox!" -c "tux"']
            }
        },
        # typo in envlist called in main.py
        {
            'wox_configuration': {'recreate_envs': False},
            'showcase': {
                'envlst': ['default'],
                'deps': ['cowsay'],
                'commands': ['cowsay -t "Hello Wox!" -c "tux"']
            }
        }
    ]

    # python -B -m pytest -rA -k "test_missing_wox_keyword_error"
    @pytest.mark.parametrize('wrong_wox_data', wrong_wox_data)
    def test_missing_wox_keyword_error(self, monkeypatch, tmp_path, capsys, wrong_wox_data):

        def patched_wox_data(self):
            return wrong_wox_data

        patched_wox_data_cached_property = cached_property(patched_wox_data)
        patched_wox_data_cached_property.__set_name__(WoxSteward, 'wox_data')

        monkeypatch.setattr(WoxSteward, 'wox_data', patched_wox_data_cached_property)
        monkeypatch.setattr('wox.main.DOT_WOX', tmp_path)

        exit_code = cli_controller({})

        assert exit_code == ExitCode.CONFIG_ERROR.value

        out = capsys.readouterr().out

        assert 'MissingWoxKeywordError' in out

# python -B -m pytest -rA -k "test_wox_session_interrupted"
def test_wox_session_interrupted(monkeypatch):
    """
        Wox session interrupted must return an INTERRUPTED exit code.
    """

    wox_toml_path = Path('tests/resources/config-files/nominal-run/wox.toml')

    monkeypatch.setattr(WoxSteward, 'wox_toml', wox_toml_path)

    def interruption(*args, **kwargs):
        raise KeyboardInterrupt

    monkeypatch.setattr('wox.configurator.configuration_creation', interruption)

    exit_code = cli_controller({})

    assert exit_code == ExitCode.INTERRUPTED.value

# python -B -m pytest -rA -k "test_wox_session_default_setting"
def test_wox_session_default_setting(monkeypatch, caplog):
    """
        'wox' command must only run the tasks listed in wox_session_default.
    """

    wox_toml_path = Path(
        'tests/resources/config-files/wox_session_default-setting/wox.toml'
    )

    monkeypatch.setattr(WoxSteward, 'wox_toml', wox_toml_path)

    def interruption(*args, **kwargs):
        raise KeyboardInterrupt

    monkeypatch.setattr('wox.venv_management.venv_deletion', interruption)

    exit_code = cli_controller({'verbose': True})

    assert exit_code == ExitCode.INTERRUPTED.value

    wox_configuration = extract_dict_from_caplog(caplog, '- wox configuration')

    wox_data = extract_dict_from_caplog(caplog, '- task settings')

    assert wox_configuration['wox_session_default'] == list(wox_data)

# python -B -m pytest -rA -k "test_task_priority_over_wox_session_default_setting"
def test_task_priority_over_wox_session_default_setting(monkeypatch, caplog):
    """
        'wox --task' command must be priority over wox_session_default.
    """

    task = 'doc'

    wox_toml_path = Path(
        'tests/resources/config-files/wox_session_default-setting/wox.toml'
    )

    monkeypatch.setattr(WoxSteward, 'wox_toml', wox_toml_path)

    def interruption(*args, **kwargs):
        raise KeyboardInterrupt

    monkeypatch.setattr('wox.venv_management.venv_deletion', interruption)

    exit_code = cli_controller({'verbose': True, 'task': [task]})

    assert exit_code == ExitCode.INTERRUPTED.value

    wox_configuration = extract_dict_from_caplog(caplog, '- wox configuration')

    wox_data = extract_dict_from_caplog(caplog, '- task settings')

    assert task in wox_data
    assert task not in wox_configuration['wox_session_default']

# python -B -m pytest -rA -k "test_dependency_conflict_in_deps"
def test_dependency_conflict_in_deps(monkeypatch, tmp_path, caplog):
    """
        Wox session must return a RUNTIME_ERROR exit code due to
        DependencyConflictError.
    """

    wox_toml_path = Path(
        'tests/resources/config-files/dependency-conflict-in-deps/wox.toml'
    )

    monkeypatch.setattr(WoxSteward, 'wox_toml', wox_toml_path)
    monkeypatch.setattr('wox.main.DOT_WOX', tmp_path)

    exit_code = cli_controller({})

    assert exit_code == ExitCode.RUNTIME_ERROR.value

    log_record_message = caplog.records[-1].msg

    assert 'DependencyConflictError' == log_record_message.type

# python -B -m pytest -rA -k "test_command_not_found"
def test_command_not_found(monkeypatch, tmp_path, caplog):
    """
        Wox session must return a RUNTIME_ERROR exit code due to
        CommandNotFoundError.
    """

    wox_toml_path = Path('tests/resources/config-files/command-not-found/wox.toml')

    monkeypatch.setattr(WoxSteward, 'wox_toml', wox_toml_path)
    monkeypatch.setattr('wox.main.DOT_WOX', tmp_path)

    exit_code = cli_controller({})

    assert exit_code == ExitCode.RUNTIME_ERROR.value

    log_record_message = caplog.records[-1].msg

    assert 'CommandNotFoundError' == log_record_message.type

# python -B -m pytest -rA -k "test_python_interpreter_call"
def test_python_interpreter_call(monkeypatch, tmp_path, caplog):
    """
        Wox session must return a RUNTIME_ERROR exit code due to
        PythonInterpreterCallError.
    """

    wox_toml_path = Path(
        'tests/resources/config-files/python-interpreter-call/wox.toml'
    )

    monkeypatch.setattr(WoxSteward, 'wox_toml', wox_toml_path)
    monkeypatch.setattr('wox.main.DOT_WOX', tmp_path)

    exit_code = cli_controller({})

    assert exit_code == ExitCode.RUNTIME_ERROR.value

    log_record_message = caplog.records[-1].msg

    assert 'PythonInterpreterCallError' == log_record_message.type

# python -B -m pytest -rA -k "test_python_not_found"
def test_python_not_found(monkeypatch, tmp_path, caplog):
    """
        Wox session must return a RUNTIME_ERROR exit code due to
        PythonInterpreterNotFoundError.
    """

    wox_toml_path = Path('tests/resources/config-files/python-not-found/wox.toml')

    monkeypatch.setattr(WoxSteward, 'wox_toml', wox_toml_path)
    monkeypatch.setattr('wox.main.DOT_WOX', tmp_path)

    exit_code = cli_controller({})

    assert exit_code == ExitCode.RUNTIME_ERROR.value

    log_record_message = caplog.records[-1].msg

    assert 'PythonInterpreterNotFoundError' == log_record_message.type

# python -B -m pytest -rA -k "test_python_build_keyword"
@maintainer_only
def test_python_build_keyword(monkeypatch, tmp_path, caplog, subtests):
    """
        Wox must discover Python interpreter for versions 3.10, 3.12, 3.13 and 3.14 and
        use the python_build provided for versions 3.11 and 3.13.
    """

    expected_logs = [
        "Looking for an interpreter version 3.10",
        "Use of specified python build '~/python313/bin/python3'",
        "Looking for an interpreter version 3.12",
        "Use of specified python build '~/python311/bin/python3'"
    ]

    caplog.set_level('DEBUG')

    wox_toml_path = Path('tests/resources/config-files/python_build-keyword/wox.toml')

    monkeypatch.setattr(WoxSteward, 'wox_toml', wox_toml_path)
    monkeypatch.setattr('wox.main.DOT_WOX', tmp_path)

    exit_code = cli_controller({'verbose': True})

    assert exit_code == ExitCode.SUCCESS.value

    for log in expected_logs:
        with subtests.test(log = log):
            assert log in caplog.text

# python -B -m pytest -rA -k "test_python_build_typo"
def test_python_build_typo(monkeypatch, tmp_path, caplog):
    """
        Wox must return PythonBuildNotFoundError if there is a typo
        in python_build setting in .toml file.
    """

    wox_toml_path = Path('tests/resources/config-files/python_build-typo/wox.toml')

    monkeypatch.setattr(WoxSteward, 'wox_toml', wox_toml_path)
    monkeypatch.setattr('wox.main.DOT_WOX', tmp_path)

    exit_code = cli_controller({'verbose': True})

    assert exit_code == ExitCode.RUNTIME_ERROR.value

    log_record_message = caplog.records[-1].msg

    assert 'PythonBuildNotFoundError' == log_record_message.type

# python -B -m pytest -rA -k "test_subprocess_traceback_display"
def test_subprocess_traceback_display(monkeypatch, tmp_path, caplog):
    """
        If failing, the subprocess must show a traceback as it would
        be shown in the terminal.
    """

    expected_lines_in_traceback = [
        'Traceback (most recent call last):',
        "NameError: name 'a' is not defined"
    ]

    wox_toml_path = Path('tests/resources/config-files/traceback-display/wox.toml')

    monkeypatch.setattr(WoxSteward, 'wox_toml', wox_toml_path)
    monkeypatch.setattr('wox.main.DOT_WOX', tmp_path)

    exit_code = cli_controller({})

    assert exit_code == ExitCode.RUNTIME_ERROR.value

    traceback_from_logs = extract_traceback_from_caplog(caplog.text)

    lines_in_traceback = []

    for line in expected_lines_in_traceback:
        if line in traceback_from_logs:
            lines_in_traceback.append(line)

    assert lines_in_traceback == expected_lines_in_traceback

# python -B -m pytest -rA -k "test_wrong_toml_file"
def test_wrong_toml_file(monkeypatch, tmp_path, capsys):
    """
        Wox must return PythonInterpreterNotFoundError if there is a typo
        in python_build setting in .toml file.
    """

    wox_toml_path = Path('tests/resources/config-files/wrong-toml-file/wox.toml')

    monkeypatch.setattr(WoxSteward, 'wox_toml', wox_toml_path)
    monkeypatch.setattr('wox.main.DOT_WOX', tmp_path)

    exit_code = cli_controller({'verbose': True})

    assert exit_code == ExitCode.CONFIG_ERROR.value

    out = capsys.readouterr().out

    assert 'TOMLDecodeError' in out

# python -B -m pytest -rA -k "test_unstable_python_py36_with_uv"
@maintainer_only
def test_unstable_python_py36_with_uv(monkeypatch, tmp_path, caplog):
    """
        Wox with Python 3.6 and ``with_uv`` setting declared must use venv
        for environment creation and pip for packages installation.

        OR

        Wox running with a corrupted Python must throw PythonInterpreterUnstableError.
    """

    wox_toml_path = Path('tests/resources/config-files/python-unstable/wox.toml')

    monkeypatch.setattr(WoxSteward, 'wox_toml', wox_toml_path)
    monkeypatch.setattr('wox.main.DOT_WOX', tmp_path)

    exit_code = cli_controller({'verbose': True})

    if exit_code == ExitCode.SUCCESS.value:

        assert exit_code == ExitCode.SUCCESS.value

    elif exit_code == ExitCode.RUNTIME_ERROR.value:

        log_record_message = caplog.records[-1].msg

        assert 'PythonInterpreterUnstableError' == log_record_message.type

# python -B -m pytest -rA -k "test_no_task_in_toml_file"
def test_no_task_in_toml_file(monkeypatch, tmp_path, capsys):
    """
        Wox must return NoTaskError if there is no tasks in .toml file.
    """

    wox_toml_path = Path('tests/resources/config-files/no-task-in-toml/wox.toml')

    monkeypatch.setattr(WoxSteward, 'wox_toml', wox_toml_path)
    monkeypatch.setattr('wox.main.DOT_WOX', tmp_path)

    exit_code = cli_controller({})

    assert exit_code == ExitCode.CONFIG_ERROR.value

    out = capsys.readouterr().out

    assert 'NoTaskError' in out

# python -B -m pytest -rA -k "test_no_python_venv_warning"
def test_no_python_venv_warning(monkeypatch, tmp_path, caplog):
    """
        If an environment is declared in an envlist and does not have pyvenv.cfg,
        wox must consider it as a non-Python environment and warn the user
        thanks to NotPythonVenvWarning.
    """

    env_dir = tmp_path / 'project'
    env_dir.mkdir()

    wox_toml_path = Path('tests/resources/config-files/not-python-venv-warning/wox.toml')

    monkeypatch.setattr(WoxSteward, 'wox_toml', wox_toml_path)
    monkeypatch.setattr('wox.main.DOT_WOX', tmp_path)

    exit_code = cli_controller({})

    assert exit_code == ExitCode.SUCCESS.value

    assert 'NotPythonVenvWarning' in caplog.text



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'wox'
copyright = '2025, wbarillon'
author = 'wbarillon'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'sphinx_design'
]

templates_path = ['_templates']
exclude_patterns = []

add_module_names = False

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'furo'
html_static_path = ['_static']
html_css_files = ['css/custom.css', 'css/colors.css']
html_js_files = ['js/custom.js']

# Use of already included s5defs.txt
# https://docutils.sourceforge.io/docs/ref/rst/definitions.html
# https://docutils.sourceforge.io/docutils/parsers/rst/include/s5defs.txt
# Enables the uses of role :color: (example at command_line_options.rst)
rst_prolog = """
.. include:: <s5defs.txt>

"""



import logging

from wox.logger.formatters.colored_formatter import colored_formatter


def initialize_console_handler():
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(colored_formatter)

    return console_handler



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
import sys
from pathlib import Path


BASE_DIR = Path(__file__).parents[3]
sys.path.append(str(BASE_DIR))


# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'Tests'
copyright = '2025, wbarillon'
author = 'wbarillon'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode'
]

templates_path = ['_templates']
exclude_patterns = []

add_module_names = False

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'furo'
html_static_path = ['_static']



CONTEXTUAL_INFO = {
    'task': 'wox',
    'env': 'host'
}

DATEFMT = '%Y-%m-%d %H:%M:%S'

# f-strings style sets to ease the use of padding. See https://medium.com/@johnidouglasmarangon/padding-f-strings-in-python-977b17edbd36
STYLE = '{'



# coverage run -m pytest -rA tests/test_subprocessor_error_detection.py --no-header --html=./reports/pytest/report.html --self-contained-html && coverage html #noqa E501

import os
import stat
import sys
import platform
import pytest

from pathlib import Path

import wox.python_discoverer as python_discoverer
import wox.subprocessor as subprocessor
import wox.venv_management as venv_manager

from wox.wox_exceptions import (
    DependencyConflictError,
    DiskFullError,
    PackageNotFoundError,
    PermissionDeniedError,
    PipTimeoutError,
    PythonInterpreterNotFoundError,
    PythonInterpreterUnstableError
)
from tests.conftest import maintainer_only
from tests.resources.constants import RAMDISK_PATH
from tests.resources.utils import run_with_timeout_process


@pytest.fixture
def package():
    return 'pandas'

@pytest.fixture
def python_interpreter():
    return python_discoverer.PythonInterpreter(
        python_discoverer.Version(platform.python_version()),
        Path(sys.executable)
    )

# python -B -m pytest -rA -k "test_pip_dependency_conflict_error"
def test_pip_dependency_conflict_error(tmp_path, python_interpreter):

    with pytest.raises(DependencyConflictError):
        subprocessor.run_command(
            f'{python_interpreter.path} -I -m pip install -qqq \
            --target {tmp_path} sphinx==8.2.3 furo==2021.10.9',
            python_interpreter = python_interpreter
        )

# python -B -m pytest -rA -k "test_uv_dependency_conflict_error"
def test_uv_dependency_conflict_error(tmp_path, python_interpreter):

    with pytest.raises(DependencyConflictError):
        subprocessor.run_command(
            f'uv pip install --python {python_interpreter.path} -qqq \
            --target {tmp_path} sphinx==8.2.3 furo==2021.10.9',
            python_interpreter = python_interpreter
        )

# python -B -m pytest -rA -k "test_uv_disk_full"
def test_uv_disk_full(tmpfs, python_interpreter):

    with pytest.raises(DiskFullError):
        subprocessor.run_command(
            f'uv pip install --python {python_interpreter.path} -qqq \
            --target {RAMDISK_PATH} pandas',
            python_interpreter = python_interpreter
        )

# python -B -m pytest -rA -k "test_pip_disk_full"
@pytest.mark.flaky(reruns = 1, reruns_delay = 1)
def test_pip_disk_full(tmpfs, python_interpreter):

    def pip_install(python_interpreter):
        subprocessor.run_command(
            f'{python_interpreter.path} -I -m pip install -qqq --target {RAMDISK_PATH} \
            pandas',
            python_interpreter = python_interpreter
        )

    with pytest.raises(DiskFullError):
        run_with_timeout_process(pip_install,
            timeout = 10,
            python_interpreter = python_interpreter
        )

# python -B -m pytest -rA -k "test_pip_package_not_found_error"
def test_pip_package_not_found_error(tmp_path, python_interpreter):

    with pytest.raises(PackageNotFoundError):
        subprocessor.run_command(
            f'{python_interpreter.path} -I -m pip install -qqq --target {tmp_path} pytes \
            cowsaa',
            python_interpreter = python_interpreter
        )

# python -B -m pytest -rA -k "test_uv_package_not_found_error"
def test_uv_package_not_found_error(tmp_path, python_interpreter):

    with pytest.raises(PackageNotFoundError):
        subprocessor.run_command(
            f'uv pip install --python {python_interpreter.path} -qqq \
            --target {tmp_path} pytes cowsaa',
            python_interpreter = python_interpreter
        )

# python -B -m pytest -rA -k "test_pip_permission_denied_error"
@maintainer_only
def test_pip_permission_denied_error(tmp_path, package, python_interpreter):
    print(os.environ.get('RUN_MAINTAINER_TESTS'))
    os.chmod(tmp_path, stat.S_IREAD)

    with pytest.raises(PermissionDeniedError):
        subprocessor.run_command(
            f'{python_interpreter.path} -I -m pip install -qqq --target {tmp_path} \
            {package}',
            python_interpreter = python_interpreter
        )

# python -B -m pytest -rA -k "test_uv_permission_denied_error"
@maintainer_only
def test_uv_permission_denied_error(tmp_path, package, python_interpreter):

    os.chmod(tmp_path, stat.S_IREAD)

    with pytest.raises(PermissionDeniedError):
        subprocessor.run_command(
            f'uv pip install --python {python_interpreter.path} -qqq --target {tmp_path} \
            {package}',
            python_interpreter = python_interpreter
        )

# python -B -m pytest -rA -k "test_pip_timeout_error"
@pytest.mark.parametrize(
    'url', [
        'http://127.0.0.1:9999/simple',
        'https://192.0.2.1/simple'
    ],
    ids = [
        'localhost',
        'reserved_ip'
    ]
)
def test_pip_timeout_error(tmp_path, package, python_interpreter, url):
    """
        Simulates a network failure when pip cannot reach its index server.
        Uses an IP address in the 192.0.2.0/24 range, reserved for
        documentation and testing purposes.
    """

    with pytest.raises(PipTimeoutError):
        subprocessor.run_command(
            f"{python_interpreter.path} -I -m pip install -qqq \
            --target {tmp_path} --cache-dir {tmp_path / 'pip_cache'} --index-url {url} \
            {package}",
            python_interpreter = python_interpreter
        )

# python -B -m pytest -rA -k "test_uv_timeout_error"
@pytest.mark.parametrize(
    'url', [
        'http://127.0.0.1:9999/simple',
        'https://192.0.2.1/simple'
    ],
    ids = [
        'localhost',
        'reserved_ip'
    ]
)
def test_uv_timeout_error(tmp_path, package, python_interpreter, url):
    """
        Simulates a network failure when pip cannot reach its index server.
        Uses an IP address in the 192.0.2.0/24 range, reserved for
        documentation and testing purposes.
    """

    with pytest.raises(PipTimeoutError):
        subprocessor.run_command(
            f"uv pip install --python {python_interpreter.path} -qqq \
            --target {tmp_path} --cache-dir {tmp_path / 'uv_cache'} --index {url} \
            {package}",
            python_interpreter = python_interpreter
        )

# python -B -m pytest -rA -k "test_pip_python_interpreter_unstable_error"
@maintainer_only
def test_pip_python_interpreter_unstable_error(tmp_path):

    if platform.system() != 'Linux':
        pytest.skip(
            'Requires Python broken interpreter built from source' \
            'following this guide : <url to doc>.'
        )

    env = 'py311'

    python_interpreter = python_discoverer.PythonInterpreter(
        python_discoverer.Version('3.11.3'),
        Path.home() / 'python311-broken' / 'bin' / 'python3.11'
    )

    env_path = tmp_path / env

    with pytest.raises(PythonInterpreterUnstableError):
        venv_manager.create_venv(env_path, python_interpreter)

# python -B -m pytest -rA -k "test_uv_python_interpreter_unstable_error"
@maintainer_only
def test_uv_python_interpreter_unstable_error(tmp_path):

    if platform.system() != 'Linux':
        pytest.skip('only works on Linux')

    if platform.system() != 'Linux':
        pytest.skip(
            'Requires Python broken interpreter built from source' \
            'following this guide : <url to doc>.'
        )

    env = 'py311'

    python_interpreter = python_discoverer.PythonInterpreter(
        python_discoverer.Version('3.11.3'),
        Path.home() / 'python311-broken' / 'bin' / 'python3.11'
    )

    env_path = tmp_path / env

    with pytest.raises(PythonInterpreterUnstableError):
        venv_manager.create_venv(env_path, python_interpreter, with_uv = True)

# python -B -m pytest -rA -k "test_python_interpretor_not_found_error"
def test_python_interpretor_not_found_error():

    with pytest.raises(PythonInterpreterNotFoundError):
        python_discoverer.python_discovery('py30')



# coverage run -m pytest -rA tests/test_cli.py --no-header --html=./reports/pytest/report.html --self-contained-html && coverage html #noqa E501

import pytest
import sys

from pathlib import Path

from wox.cli import main_cli
from wox.cli_control import cli_controller
from wox.constants import ExitCode
from wox.wox_steward import WoxSteward
from wox._version import __version__


# python -B -m pytest -rA -k "test_task_not_found_error_with_task"
def test_task_not_found_error_with_task(monkeypatch, capsys):
    """
        A task passed in ``wox --task`` not in .toml file must return TaskNotFoundError.
    """

    wox_toml_path = Path('tests/resources/config-files/nominal-run/wox.toml')

    monkeypatch.setattr(WoxSteward, 'wox_toml', wox_toml_path)
    monkeypatch.setattr(sys, 'argv', ['wox', '--task', 'pixegami', 'cowsa', 'bash'])

    with pytest.raises(SystemExit) as e:
        main_cli()

    assert e.value.code == ExitCode.CONFIG_ERROR.value

    out = capsys.readouterr().out

    assert 'TaskNotFoundError' in out

# python -B -m pytest -rA -k "test_task_not_found_error_with_exclude_task"
def test_task_not_found_error_with_exclude_task(monkeypatch, capsys):
    """
        A task passed in ``wox --task`` not in .toml file must return TaskNotFoundError.
    """

    wox_toml_path = Path('tests/resources/config-files/nominal-run/wox.toml')

    monkeypatch.setattr(WoxSteward, 'wox_toml', wox_toml_path)
    monkeypatch.setattr(sys, 'argv', ['wox', '--exclude-task', 'pixegami', 'cowsa', 'bash'])

    with pytest.raises(SystemExit) as e:
        main_cli()

    assert e.value.code == ExitCode.CONFIG_ERROR.value

    out = capsys.readouterr().out

    assert 'TaskNotFoundError' in out

# python -B -m pytest -rA -k "test_wox_task_parsing"
def test_wox_task_parsing(monkeypatch):
    """
        ``wox --task`` must be able to have as many tasks as necessary.
    """

    control_cli_options = ['pixegami', 'cowsay', 'bash']

    def patched_cli_controller(cli_options):
        patched_cli_controller.cli_options = cli_options
        return ExitCode.SUCCESS.value

    monkeypatch.setattr(sys, 'argv', ['wox', '--task', 'pixegami', 'cowsay', 'bash'])
    monkeypatch.setattr('wox.cli_control.cli_controller', patched_cli_controller)

    with pytest.raises(SystemExit) as e:
        main_cli()

    assert e.value.code == ExitCode.SUCCESS.value

    assert patched_cli_controller.cli_options['task'] == control_cli_options

# python -B -m pytest -rA -k "test_wox_verbosity"
def test_wox_verbosity(monkeypatch):
    """
        Wox verbosity must be boolean and not cumulative.
    """

    def patched_cli_controller(cli_options):
        patched_cli_controller.cli_options = cli_options
        return ExitCode.SUCCESS.value

    monkeypatch.setattr(sys,
        'argv', ['wox', '--verbose', '--task', 'pixegami', 'cowsay', 'bash']
    )
    monkeypatch.setattr('wox.cli_control.cli_controller', patched_cli_controller)

    with pytest.raises(SystemExit) as e:
        main_cli()

    assert e.value.code == ExitCode.SUCCESS.value

    assert isinstance(patched_cli_controller.cli_options['verbose'], bool)
    assert patched_cli_controller.cli_options['verbose'] is True

# python -B -m pytest -rA -k "test_wox_version"
@pytest.mark.parametrize('option', ['-V', '--version'])
def test_wox_version(monkeypatch, capsys, option):
    """
        ``wox --version`` (or -V) must return the version of wox.
    """

    monkeypatch.setattr(sys, 'argv', ['wox', option])

    with pytest.raises(SystemExit) as e:
        main_cli()

    assert e.value.code == ExitCode.SUCCESS.value

    out = capsys.readouterr().out

    assert __version__ in out

# python -B -m pytest -rA -k "test_wox_command_without_config_file"
def test_wox_without_config_file(monkeypatch, tmp_path, cwd_tmp_path, patched_wox_command):
    """
        Wox command run without configuration file must delete .wox directory.
    """

    tmp_dot_wox = tmp_path / '.wox'
    tmp_dot_wox.mkdir()

    monkeypatch.setattr('wox.cli_control.DOT_WOX', tmp_dot_wox)

    with pytest.raises(SystemExit):
        main_cli()

    assert not tmp_dot_wox.exists()

# python -B -m pytest -rA -k "test_wox_purge"
def test_wox_purge(monkeypatch, tmp_path):
    """
        ``wox --purge`` must delete all directories and files excepting .logs directory.
        At the end of the purge, .wox must have .logs directory, .gitignore file
        and CACHEDIR.TAG file.
    """

    dot_wox = tmp_path / '.wox'

    directory = dot_wox / '.venv'
    directory.mkdir(parents = True)

    file = dot_wox / 'file.txt'
    file.touch()

    logs_dir = dot_wox / '.logs'
    logs_dir.mkdir(parents = True)

    # Keeps the inode (file or directory identifier)
    original_logs_inode = logs_dir.stat().st_ino

    monkeypatch.setattr('wox.wox_steward.DOT_WOX', dot_wox)

    exit_code = cli_controller({'purge': True})

    assert exit_code == ExitCode.SUCCESS.value

    # Ensures that the .logs directory is exactly the same one (same inode),
    # not deleted and recreated during the purge.
    assert logs_dir.stat().st_ino == original_logs_inode

    remaining = sorted(path.name for path in (tmp_path / '.wox').iterdir())

    assert remaining == ['.gitignore', '.logs', 'CACHEDIR.TAG']



# coverage run -m pytest -rA tests/test_logging.py --no-header --html=./reports/pytest/report.html --self-contained-html && coverage html #noqa E501

from wox.logger.logger_setup import wox_logger


# python -B -m pytest -rA -k "test_logging"
def test_logging(caplog):
    """
        The logger must correctly handle non-ASCII characters.
    """

    caplog.set_level('INFO')

    wox_logger.info('㐬')

    # caplog.records make logs accessible for assertion.
    log_record_message = caplog.records[0].msg

    assert log_record_message == '㐬'



"""
    This module contains all error that may occur during the management of
    virtual environments handled by Python3's built-in solutions (venv and pip).
"""

import re

from wox.subprocess_error_detection.utils import detect_error_from_stderr
from wox.utils import get_python_version_from_command
from wox.wox_exceptions import (
    DependencyConflictError,
    DiskFullError,
    PackageNotFoundError,
    PermissionDeniedError,
    PipTimeoutError,
    PythonInterpreterUnstableError
)


def detect_pip_errors(stderr_output, command, python_interpreter):
    """
        Detects pip-related errors upon dependencies installation.

        Parameters
        ----------
        stderr_output : str
            Output captured from the subprocess stderr stream.
        command : str
            Command executed by subprocess.
        python_interpreter : PythonInterpreter
            Python interpreter with its version and path.

        Raises
        ------
        DependencyConflictError
            Dependency conflict occurs during package installation.
        DiskFullError
            Disk is full and command cannot write data.
        PipTimeoutError
            Pip cannot reach package server.
        PackageNotFoundError
            Package is missing or wrongly spelled.
        PermissionDeniedError
            Permissions prevent writing data.
        PythonInterpreterUnstableError
            The Python interpreter crashes at the C-level.
    """

    check_dependency_conflict_error(stderr_output, command)
    check_disk_full_error(stderr_output, command)
    check_pip_timeout_error(stderr_output, command)
    check_package_not_found_error(stderr_output, command, python_interpreter.path)
    check_permission_denied_error(stderr_output, command)
    check_python_interpreter_unstable_error(stderr_output,
        command,
        python_interpreter.version
    )

def check_dependency_conflict_error(stderr_output, command):
    if (
        'install' in command and \
        detect_error_from_stderr(stderr_output,
            ['resolutionimpossible', 'conflicting dependencies']
        )
    ):
        raise DependencyConflictError

def check_disk_full_error(stderr_output, command):
    if 'install' in command and detect_error_from_stderr(stderr_output, ['errno 28']):
        raise DiskFullError

def check_pip_timeout_error(stderr_output, command):
    if ('install' in command and \
        detect_error_from_stderr(stderr_output,
            ['connection broken', 'failed to establish new connection']
        )
    ):
        raise PipTimeoutError

def check_package_not_found_error(stderr_output, command, python_interpreter_path = None):
    if ('install' in command and \
        detect_error_from_stderr(stderr_output,
            ['could not find a version', 'no matching distribution']
        )
    ):
        environment = python_interpreter_path.parent.parent.name
        match = re.search(r'No matching distribution found for (\S+)', stderr_output)
        if match:
            missing_package = match.group(1)
        raise PackageNotFoundError(missing_package, environment)

def check_permission_denied_error(stderr_output, command):
    if ('install' in command and \
        detect_error_from_stderr(stderr_output, ['permissionerror'])
    ):
        raise PermissionDeniedError

def check_python_interpreter_unstable_error(stderr_output, command, version):
    if (
        (
            ('venv' in command or 'install' in command) and \
            detect_error_from_stderr(stderr_output,
                [
                    'died with', 'sigsegv',
                    'segmentation fault', 'core dumped',
                    'problem confirming the ssl certificate',
                    'ssl module is not available',
                    'ssl module in python is not available'
                ]
            )
        ) or (
            'cannot open shared object file' in stderr_output.lower() and \
            'libpython' in stderr_output.lower()
        )
    ):
        if not version:
            version = get_python_version_from_command(command)
        raise PythonInterpreterUnstableError(version)



import os
import re
import shlex
import shutil
import time

from wox.constants import DOT_WOX


def auto_clean_logs():
    """
        Setting up logging for wox is bound to create a log file. When
        wox crash before the session starts, the log file is empty and
        pollutes the wox session log directory.

        To mitigate this flaw from logging, this function will remove
        the last log file created.
    """

    logs_dir = DOT_WOX / '.logs'

    list_of_files = list(logs_dir.glob('**/output*'))

    if list_of_files:
        latest_file = max(list_of_files, key=os.path.getctime)
        os.remove(latest_file)

def normalize_python_version(digits):
    """
        Ensures correct major.minor version scheme for Python 3 (e.g. 36 => 3.6).

        Parameters
        ----------
        digits: str

        Returns
        -------
        str
    """

    return f'{digits[0]}.{digits[1:]}'

def get_python_version_from_command(command):
    if isinstance(command, str):
        command = shlex.split(command)

    python_interpreter = command[0]

    if python_interpreter == 'py':
        version = re.findall(r'\d+', ' '.join(command))
        version = '.'.join(version)
    else:
        version = re.sub(r'[a-zA-Z]', '', python_interpreter)

    return version

def sort_set(set, iterable):
    """
        Sorts the set with order from original iterable.

        Returns
        -------
        list
    """

    return [element for element in iterable if element in set]

def format_duration(duration):
    """
        Format a duration in seconds into an ISO-like time string
        (e.g. 10 seconds => 00:00:10.XXX).

        Parameters
        ----------
        duration : float
            Duration in seconds.

        Returns
        -------
        str

        Notes
        -----
        This function uses ``divmod()`` to progressively extract hours, minutes and
        seconds from the total duration. ``divmod(a, b)`` returns a tuple
        ``(quotient, remainder)``, equivalent to ``(a // b, a % b)``.

        Examples
        --------
        Both implementations return 1 hour and 71.527 seconds:

        >>> divmod(3671.527, 3600)
        (1.0, 71.527)

        This is equivalent to:
        >>> hours = 3671.527 // 3600
        >>> remainder = 3671.527 % 3600
        >>> (hours, remainder)
        (1.0, 71.527)
    """

    hours, remaining_seconds = divmod(duration, 3600)
    minutes, remaining_seconds = divmod(remaining_seconds, 60)
    seconds = int(remaining_seconds)
    milliseconds = int((remaining_seconds - seconds) * 1000)

    return f'{int(hours):02d}:{int(minutes):02d}:{seconds:02d}.{milliseconds:03d}'

def rmtree_with_retry(path, retries = 5, delay = 0.2):
    """
        Safety layer around shutil.rmtree ensuring the deletion of virtual environment.

        Parameters
        ----------
        path : str
            Path to the virtual environment.
        retries : int
        delay : float

        Raises
        ------
        PermissionError
            shutil.rmtree was not authorized to delete the arborescence.
    """

    for attempt in range(retries):
        try:
            shutil.rmtree(path)
            return
        except PermissionError:
            if attempt + 1 == retries:
                raise
            time.sleep(delay)



import logging

from wox.logger.constants import CONTEXTUAL_INFO


class SetLevelTo(logging.Filter):
    """
        A filter mimicking the behaviour of setLevel().

        Example :
        logger.filters = [SetLevelTo('INFO')]

        Parameters
        ----------
        level : str

        Returns
        -------
        Boolean
            True if the LogRecord has level number greater or
            equal than level.
    """

    def __init__(self, level):
        self.level = level

    def filter(self, record):
        return record.levelno >= logging._nameToLevel[self.level]

class WoxContext(logging.Filter):
    """
        Adds contextual fields to log formatting.

        Parameters
        ----------
        contextual_info : dict, default wox.logger.constants.CONTEXTUAL_INFO
            The default contextual information.
    """

    def __init__(self, contextual_info = None):
        if contextual_info is None:
            self.contextual_info = CONTEXTUAL_INFO.copy()
        else:
            self.contextual_info = contextual_info

    def filter(self, record):
        """
            Returns
            -------
            Boolean
                True if contexual info are set in LogRecord.
        """

        for key, value in self.contextual_info.items():
            setattr(record, key, value)

        return True

    def update_context(self, overrides):
        """
            Updates the current contextual information.

            Parameters
            ----------
            overrides : dict
                Key-value pairs overriding the current contextual information.
        """

        self.contextual_info.update(overrides)



from enum import IntEnum
from pathlib import Path


DOT_WOX = Path.cwd() / '.wox'

# MSV stands for minimum supported version
MSV_VIRTUALENV = '>= 3.7'
MSV_UV = '>= 3.8'

class ExitCode(IntEnum):
    """
        Standardized exit codes for Wox CLI.

        Notes
        -----
        0 - SUCCESS
            Normal execution completed successfully.

        1 - RUNTIME_ERROR
            Unhandled or internal error occurred during execution.

        2 - CONFIG_ERROR
            Misconfiguration or missing setup.

        130 - INTERRUPTED
            Process interrupted by user (SIGINT).
    """

    SUCCESS = 0
    RUNTIME_ERROR = 1
    CONFIG_ERROR = 2
    INTERRUPTED = 130



import os
import pytest

from pathlib import Path


@pytest.fixture
def pyproject_toml_path():
    return Path('tests/resources/config-files/pyproject.toml').resolve()

@pytest.fixture
def wox_toml_path():
    return Path('tests/resources/config-files/wox.toml').resolve()

@pytest.fixture
def cwd_tmp_path(tmp_path):
    """
        Path.cwd() is the root to the python execution. Files location
        is based on it. For some tests, it is useful to alter this
        path to pytest temporal directory created for the test.

        Since pytest creates temporary directories
        using tmp_path *outside* the project tree, we must temporarily
        change the working directory to tmp_path for the function to succeed.
    """

    old_cwd = os.getcwd()
    os.chdir(tmp_path)
    try:
        yield tmp_path
    finally:
        os.chdir(old_cwd)



import logging

from wox.logger.constants import CONTEXTUAL_INFO


def update_context(logger, contextual_info_overrides):
    """
        Returns a LoggerAdapter that enriches log records with contextual information.

        Parameters
        ----------
        logger : logging.Logger
            The base logger to wrap.
        contextual_info_overrides : dict
            Contextual key-value pairs overriding the defaults in CONTEXTUAL_INFO.

        Returns
        -------
        logging.LoggerAdapter
            A logger adapter bound to the provided contextual information.
    """
    context = CONTEXTUAL_INFO.copy()
    context.update(contextual_info_overrides)
    return logging.LoggerAdapter(logger, context)



"""
    Wox session can run multiple tasks within several virtual environments. To help
    the user keep track of what happens in the session, a summary is created at the end
    of each wox session.

    A summary is composed of 4 piece of information:
    - Execution unit: that is the qualification of a task running in a virtual environment;
    - Exit code: 0 if the execution unit succeed or 1 if it fails;
    - Failed at: the action running by the execution unit during the crash;
    - Failure reason: the error message.
"""

from rich.console import Console
from rich.table import Table

from wox.utils import format_duration


def render(wox_session_summary, duration):
    """
        Renders the wox summary from a list of execution units.

        Parameters
        ----------
        wox_session_summary : list[dict]
        duration : float
            Duration in seconds of the wox session.
    """

    console = Console()
    table = Table()
    table.add_column('Execution unit', justify = 'center')
    table.add_column('Exit code', justify = 'center')
    table.add_column('Failed at', justify = 'center')
    table.add_column('Failure reason', justify = 'left')

    for execution_unit in wox_session_summary:
        table.add_row(
            execution_unit['execution_unit'],
            str(execution_unit['exit_code']),
            execution_unit['failed_at'],
            execution_unit['message']
        )

    console.rule(f'Wox session summary ([bold cyan]{format_duration(duration)}[/])')
    console.print(table)

def update_execution_unit_failure(execution_unit, action, exception):
    """
        Updates the execution unit upon failure.

        Parameters
        ----------
        execution_unit : dict
            Execution unit to update.
        action : str
            The last action run by the execution unit before crash.
        exception : Exception
            The exception caught during the crash.
    """

    execution_unit.update({
        'exit_code': 1,
        'failed_at': action,
        'message': str(exception)
    })

def update_execution_unit_success(execution_unit):
    """
        Updates the execution unit upon success.

        Parameters
        ----------
        execution_unit : dict
            Execution unit to update.
    """

    execution_unit.update({
        'exit_code': 0,
        'failed_at': '-',
        'message': '-'
    })



class WoxException(Exception):
    """
        Base class for wox exceptions.

        Attributes
        ----------
        message : str
        type : str
            Name of the exception class.
    """
    def __init__(self, message):
        self.message = message
        self.type = self.__class__.__name__

    def __str__(self):
        return f'{self.type}: {self.message}'

class ConfigFileNotFoundError(WoxException):
    """
        Raised when missing wox.toml or pyproject.toml file.
    """

    def __init__(self):
        super().__init__(
            message = (
                'Provide a configuration file (wox.toml or pyproject.toml). '
                'Check <url to doc>.'
            )
        )

class CommandNotFoundError(WoxException):
    """
        Raised when a command cannot be found in the system PATH.

        Parameters
        ----------
        command : str
            The command that was not found.
    """

    def __init__(self, command):
        super().__init__(message = f'Command {command} not found.')

class DependencyConflictError(WoxException):
    """
        Raised when a dependency conflict is detected during package installation.

        Suggestion: Check the subprocess logs for conflicting packages.
    """

    def __init__(self):
        super().__init__(
            message = (
                'Dependency conflict detected. Check logs from subprocess '
                'for more details.'
            )
        )

class DiskFullError(WoxException):
    """
        Raised when an operation fails due to insufficient disk space.
    """

    def __init__(self):
        super().__init__(message = 'Installation failed: disk is full.')

class MissingWoxKeywordError(WoxException):
    """
        Raised when wox setting in toml file doesn't exist.
    """

    def __init__(self, key):
        super().__init__(message = (
                f'Wox keyword {key} not found in .toml file. '
                f'Check the documentation <how to create a wox toml file>.'
            )
        )

class MissingWoxSectionError(WoxException):
    """
        Raised when wox section not found in pyproject.toml.
    """

    def __init__(self):
        super().__init__(message = 'Wox section not found in pyproject.toml')

class NoTaskError(WoxException):
    """
        Raised when ``--task`` and ``exclude-task`` have the same tasks.
    """
    def __init__(self):
        super().__init__(
            message = 'No tasks to run.'
        )

class PackageNotFoundError(WoxException):
    """
        Raised when attempting to install a package that does not exist.

        Parameters
        ----------
        missing_package : str
            Name of the missing package.
        environment : str
            Environment where the installation was attempted.
    """

    def __init__(self, missing_package, environment):
        super().__init__(message = f'Unable to install {missing_package} in {environment}')

class PermissionDeniedError(WoxException):
    """
        Raised when a write operation fails due to insufficient permissions.
    """

    def __init__(self):
        super().__init__(message = 'Unable to write files due to insufficient permissions.')

class PipTimeoutError(WoxException):
    """
        Raised when pip cannot reach the package index due to network issues or a timeout.
    """

    def __init__(self):
        super().__init__(
            message = 'Unable to connect to the package index '
            '(network timeout or unreachable source).'
        )

class PythonBuildNotFoundError(WoxException):
    """
        Raised when no Python interpreter is found for the requested version.

        Parameters
        ----------
        path : str
            Path to declared python build in .toml file.
    """

    def __init__(self, path):
        super().__init__(message = f"Python build not found on path '{path}'.")

class PythonInterpreterCallError(WoxException):
    """
        Raised when a Python interpreter call would start an interactive session and
        block execution.

        Parameters
        ----------
        command : str
            The command that was attempted.
    """

    def __init__(self, command):
        super().__init__(
            message = (
                f"The '{command}' command would have launched an interactive "
                "session of Python and stuck the session."
            )
        )

class PythonInterpreterNotFoundError(WoxException):
    """
        Raised when no Python interpreter is found for the requested version.

        Parameters
        ----------
        version : str
            Python version requested.
    """

    def __init__(self, version):
        super().__init__(message = f'Python interpreter not found for version {version}.')

class PythonInterpreterUnstableError(WoxException):
    """
        Raised when the Python interpreter is unstable and fails to execute commands.

        Suggestion: Recompile the interpreter or use a different build.

        Notes
        -----
        https://docs.python.org/3.14/using/configure.html
    """

    def __init__(self, version):
        super().__init__(
            message = (
                f'Interpreter for Python {version} appears unstable, '
                'recompile your interpreter or use a different build.'
            )
        )

class TaskNotFoundError(WoxException):
    """
        Raised when a task not present in the .toml file is passed to 'wox command --task'.
    """

    def __init__(self, task):
        super().__init__(message = f'Tasks {task} not found in .toml file.')

class NotPythonVenvWarning(WoxException):
    """
        Raised when the environment declared in .toml file is not
        a Python virtual environment.
    """

    def __init__(self, env_path):
        super().__init__(
            message = (
                f"Wox detected '{env_path.name}' as a non-Python virtual environment. "
                "Wox will use it as-is for the session."
            )
        )



import re


def detect_error_from_stderr(stderr_output, expressions):
    """
        Searches for expressions in output of subprocess.

        Parameters
        ----------
        stderr_output : str
            Output captured from the subprocess stderr stream.
        expressions : list
            List of expressions likely to be in stderr_output.

        Returns
        -------
        bool
            True if any expression is found in stderr_output.
    """

    return any(expression in stderr_output.lower() for expression in expressions)

def detect_sentence_with_fill_in_blank(stderr_output, sentence):
    """
        Searches for a fill-in-the-blank text in output of subprocess.
        All fill-in-the-blank are represented by a wildcard (*).

        Parameters
        ----------
        stderr_output : str
            Output captured from the subprocess stderr stream.
        sentence : str
            Sentence likely to be in stderr_output.

        Returns
        -------
        bool
            True if the sentence is found in stderr_output.
    """

    stderr_output = normalize_text(stderr_output)

    sentence = re.escape(sentence).replace(r'\*', '.*?')
    regex = re.compile(sentence, re.DOTALL)
    return bool(regex.search(stderr_output))

def normalize_text(text):
    """
        Replaces all whitespaces and multiples return to line by one space and
        lowers the text.

        Parameters
        ----------
        text : str

        Returns
        -------
        str
    """

    return re.sub(r'\s+', ' ', text).strip().lower()



import tomllib

from functools import cached_property
from pathlib import Path

from wox.constants import DOT_WOX
from wox.utils import rmtree_with_retry
from wox.wox_exceptions import ConfigFileNotFoundError, MissingWoxSectionError


class WoxSteward():
    """
        Manages the woxspace (.wox) and handles the settings written in
        the .toml file.

        Attributes
        ----------
        dot_wox : pathlib.Path
            Path to .wox directory.
        wox_toml : pathlib.Path
            Path to configuration file.
        wox_data : dict
            Data extracted from configuration files.
    """

    def __init__(self):
        self.dot_wox = DOT_WOX


    @cached_property
    def wox_toml(self):
        """
            Locates the configuration file.

            Returns
            -------
            pathlib.Path
                Path to config file.

            Raises
            ------
            ConfigFileNotFoundError
                No config file found at root of workspace.
        """

        toml_file_names = ['wox', 'pyproject']

        for name in toml_file_names:
            wox_toml = Path.cwd() / f'{name}.toml'
            if wox_toml.exists():
                return wox_toml

        raise ConfigFileNotFoundError

    @cached_property
    def wox_data(self):
        """
            Data extraction from configuration file.

            Returns
            -------
            dict
                Wox data from config file.

            Raises
            ------
            MissingWoxSectionError
                Wox section in pyproject.toml not found.
        """

        with open(self.wox_toml, 'rb') as toml_file:
            data = tomllib.load(toml_file)

        match self.wox_toml.name:
            case 'pyproject.toml':
                try:
                    wox_data = data['tool']['wox']
                except KeyError:
                    raise MissingWoxSectionError
            case 'wox.toml':
                wox_data = data

        return wox_data

    def dot_wox_setup(self):
        """
            Sets up the woxspace in the .wox directory.
        """

        gitignore = self.dot_wox / '.gitignore'

        if not gitignore.exists():
            with open(f'{gitignore}', 'w', encoding = 'utf-8') as file:
                file.write('# Automatically created by wox.\n')
                file.write('*\n')

        cachedir_tag = self.dot_wox / 'CACHEDIR.TAG'

        if not cachedir_tag.exists():
            cachedir_tag.write_text(
                'Signature: 8a477f597d28d172789f06886806bc55\n'
                '# This directory is managed by Wox.\n'
                '# It contains cache data such as logs and virtual environments.\n'
                '# See: https://bford.info/cachedir/\n'
            )

    def purge_wox(self):
        """
            Deletes all directories and files in .wox except .logs and recreates
            .gitignore and CACHEDIR.TAG.
        """

        for element in self.dot_wox.iterdir():
            if element.name == '.logs':
                continue

            if element.is_dir():
                rmtree_with_retry(element)
            else:
                element.unlink()

        self.dot_wox_setup()
